from __future__ import annotations

import logging
from datetime import datetime
from typing import TYPE_CHECKING

from typing_extensions import TypedDict

from vectice.models.resource.base import Resource
from vectice.models.resource.metadata import DatasetSourceOrigin
from vectice.models.resource.metadata.column_metadata import DBColumn
from vectice.models.resource.metadata.dataframe_config import DataFrameType
from vectice.models.resource.metadata.db_metadata import DBMetadata, MetadataDB

if TYPE_CHECKING:
    from snowflake.snowpark import Session
    from snowflake.snowpark.types import StructType

    TDataFrameType = DataFrameType  # because otherwise pyright can't find DataFrameType's type
    TTable = TypedDict(
        "TTable",
        {
            "num_bytes": int,
            "rows_number": int,
            "created": datetime | None,
            "modified": datetime | None,
            "schema": StructType,
        },
    )


_logger = logging.getLogger(__name__)


class SnowflakeResource(Resource):
    """Snowflake tables resource reference wrapper.

    NOTE: **This feature is currently in Beta and is subject to change. To activate this feature please contact your sales representative.**

    This resource wraps Snowflake tables paths that you have stored in Snowflake with optional metadata and versioning.
    You pass it as an argument of your Vectice Dataset wrapper before logging it to an iteration.


    ```python
    from vectice import SnowflakeResource

    connection_parameters = {
        ...
    }
    new_session = Session.builder.configs(connection_parameters).create()

    sf_resource = SnowflakeResource(
        snowflake_client=new_session,
        paths="SNOWFLAKE_SAMPLE_DATA.TPCH_SF10.PART",
    )
    ```

    NOTE: **Vectice does not store your actual dataset. Instead, it stores your dataset's metadata, which is information about your dataset.
    These details are captured by resources tailored to the specific environment in use, such as: local (FileResource), Bigquery, S3, GCS...**
    """

    _origin = DatasetSourceOrigin.SNOWFLAKE.value

    def __init__(
        self,
        paths: str | list[str],
        dataframes: TDataFrameType | list[TDataFrameType] | None = None,
        snowflake_client: Session | None = None,
        capture_schema_only: bool = False,
        force_dataframe: bool = False,
        columns_description: dict[str, str] | str | None = None,
    ):
        """Initialize a Snowflake resource.

        Parameters:
            paths: The pahts of the resources to get.
            dataframes (Optional): The dataframes allowing vectice to optionally compute more metadata about this resource such as columns stats. (Support Pandas, Spark)
            snowflake_client (Optional): The Snowflake client to retrieve table metadatas.
            capture_schema_only (Optional): A boolean parameter indicating whether to capture only the schema or both the schema and column statistics of the dataframes.
            force_dataframe (Optional): A boolean parameter indicating whether the dataframe should be inferred from the table itself.
            columns_description (Optional): A dictionary or path to a csv file to map the column's name to a specific description. Dictionary should follow the format { "column_name": "Description", ... }

        """
        super().__init__(
            paths=paths,
            dataframes=dataframes,
            capture_schema_only=capture_schema_only,
            columns_description=columns_description,
        )
        self.snowflake_client = snowflake_client
        self.force_dataframe = force_dataframe

    def _fetch_data(self) -> dict[str, tuple[TTable | None, TDataFrameType | None]]:
        tables: dict[str, tuple[TTable | None, TDataFrameType | None]] = {}
        df_index = 0
        for path in self._paths:
            path_tables, new_df_index = self._fetch_path_data(index=df_index, path=path)
            tables.update(path_tables)
            df_index += new_df_index
        return tables

    def _fetch_path_data(
        self, index: int, path: str
    ) -> tuple[dict[str, tuple[TTable | None, TDataFrameType | None]], int]:
        return self._fetch_table_data(index=index, path=path), 1

    def _fetch_table_data(self, index: int, path: str) -> dict[str, tuple[TTable | None, TDataFrameType | None]]:
        dataframe: TDataFrameType | None = (
            self._dataframes[index] if self._dataframes is not None and len(self._dataframes) > index else None
        )
        if self.snowflake_client is None:
            return {path: (None, dataframe)}

        from snowflake.snowpark.exceptions import SnowparkClientException

        tb: TTable | None = None
        try:
            schema = []
            session = self.snowflake_client
            table = session.table(path)
            schema = table.schema
            if self.force_dataframe:
                dataframe = table.to_pandas()  # pyright: ignore[reportGeneralTypeIssues]
            table_schema: str = path.split(".")[1]
            table_name = path.rpartition(".")[-1]
            sql = "SELECT ROW_COUNT, BYTES, CREATED, LAST_ALTERED FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_SCHEMA = ? and TABLE_NAME = ?"
            crsr = session.sql(sql, params=[table_schema, table_name]).collect()
            for a, b, c, d in crsr:
                tb = {"num_bytes": b, "rows_number": a, "created": c, "modified": d, "schema": schema}
        except SnowparkClientException:
            _logger.warning(f"Failed to fetch data information for table {path}")
            pass

        return {path: (tb, dataframe)}

    def _build_metadata(self) -> DBMetadata:
        tables_metadata: dict[str, tuple[TTable | None, TDataFrameType | None]] = self.data
        dbs: list[MetadataDB] = []
        for table_path, [table_metadata, dataframe] in tables_metadata.items():
            columns: list[DBColumn] | None = None
            size = None
            rows_number = None
            updated_date = None
            created_date = None
            if table_metadata is not None:
                columns = []
                size = table_metadata["num_bytes"]
                rows_number = table_metadata["rows_number"]
                modified = table_metadata["modified"]
                updated_date = modified.isoformat() if modified is not None else None
                created = table_metadata["created"]
                created_date = created.isoformat() if created is not None else None
                for column in table_metadata["schema"]:
                    columns.append(
                        DBColumn(
                            name=column.name,
                            data_type=str(column.datatype),
                            is_unique=False,
                            nullable=column.nullable,
                            is_private_key=False,
                            is_foreign_key=False,
                        )
                    )

            dbs.append(
                MetadataDB(
                    name=table_path,
                    rows_number=rows_number,
                    size=size,
                    columns=columns,
                    created_date=created_date,
                    updated_date=updated_date,
                    uri=table_path,
                    dataframe=dataframe,
                    display_name=table_path.rpartition(".")[-1],
                    capture_schema_only=self.capture_schema_only,
                )
            )

        metadata_size = None
        for db in dbs:
            if db.size is not None:
                if metadata_size is None:
                    metadata_size = 0
                metadata_size += db.size

        metadata = DBMetadata(
            size=int(metadata_size) if metadata_size else None,
            origin=self._origin,
            dbs=dbs,
        )
        return metadata



from __future__ import annotations

from dataclasses import asdict, dataclass
from datetime import datetime
from typing import Any


@dataclass
class GitVersion:
    repositoryName: str
    branchName: str
    commitHash: str
    isDirty: bool
    uri: str
    commitComment: str | None = None
    commitAuthorName: str | None = None
    commitAuthorEmail: str | None = None
    entrypoint: str | None = None

    def asdict(self) -> dict[str, Any]:
        return asdict(self)


@dataclass
class DatabricksVersion:
    url_notebook: str
    relative_path: str
    timestamp: datetime

    def asdict(self) -> dict[str, Any]:
        return {
            "urlNotebook": self.url_notebook,
            "relativePath": self.relative_path,
            "timestamp": self.timestamp.isoformat(),
        }


@dataclass
class CodeVersion:
    git_version: GitVersion | None = None
    databricks_version: DatabricksVersion | None = None

    def asdict(self) -> dict[str, Any]:
        return {
            "gitVersion": self.git_version.asdict() if self.git_version else None,
            "databricksVersion": self.databricks_version.asdict() if self.databricks_version else None,
        }



from __future__ import annotations

from typing import Any

from vectice.autolog.asset_services.metric_service import MetricService
from vectice.autolog.asset_services.property_service import PropertyService
from vectice.autolog.asset_services.technique_service import TechniqueService
from vectice.autolog.model_library import ModelLibrary


class AutologSklearnService(MetricService, PropertyService, TechniqueService):
    def __init__(self, key: str, asset: Any, data: dict):
        self._asset = asset
        self._key = key

        super().__init__(cell_data=data)

    def get_asset(self):
        # xgboost relies on BaseEstimator
        # lightgbm has Booster and sklearn API which uses BaseEstimator
        try:
            from sklearn.base import is_classifier, is_regressor
            from sklearn.pipeline import Pipeline

            if is_regressor(self._asset) or is_classifier(self._asset) or isinstance(self._asset, Pipeline):
                library = ModelLibrary.SKLEARN

                if isinstance(self._asset, Pipeline):
                    library = ModelLibrary.SKLEARN_PIPELINE
                elif str(self._asset.__class__.__module__) == "sklearn.model_selection._search":
                    library = ModelLibrary.SKLEARN_PIPELINE
                try:
                    # TODO fix regex picking up classes
                    # Ignore Initialized variables e.g LogisticRegression Class
                    self._asset.get_params()  # pyright: ignore[reportGeneralTypeIssues]
                    _, params = self._get_sklearn_or_xgboost_or_lgbm_info(self._asset)
                    return {
                        "variable": self._key,
                        "model": self._asset,
                        "library": library,
                        "technique": self._get_model_technique(self._asset, library),
                        "metrics": self._get_model_metrics(self._cell_data),
                        "properties": params,
                    }
                except Exception:
                    pass
        except ImportError:
            pass



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any, Dict, Tuple

from vectice.api.json.code_version import CodeVersion, DatabricksVersion, GitVersion
from vectice.utils.common_utils import hide_logs

if TYPE_CHECKING:
    from git.repo import Repo


_logger = logging.getLogger(__name__)


def capture_code_version() -> CodeVersion | None:
    """Capture code version.

    Implementation that captures the code version to be sent to the BE as part of registering assets for the lineage

    Returns:
        CodeVersion or None
    """
    repository, databricks_code_source = _look_for_repos()
    code_version: CodeVersion | None = None
    if repository:
        git_version: GitVersion | None = _extract_git_version(repository)
        code_version = CodeVersion(git_version=git_version) if git_version else code_version
    elif databricks_code_source is not None:
        databricks_version = _extract_databricks_code_version(databricks_code_source)
        code_version = CodeVersion(databricks_version=databricks_version) if databricks_version else code_version

    return code_version


def is_databricks():
    return "DATABRICKS_RUNTIME_VERSION" in os.environ


def _look_for_repos() -> Tuple[Repo | None, Dict[str, Any] | None]:
    databricks_code_source: Dict[str, Any] | None = None
    repository = _look_for_git_repository()

    if is_databricks():
        databricks_code_source = _look_for_databricks_code_source_info()

    return repository, databricks_code_source


def _look_for_databricks_code_source_info() -> Dict[str, Any] | None:
    databricks_code_source: Dict[str, Any] | None = None
    try:
        from IPython.core.getipython import get_ipython

        ipython_content = get_ipython().get_parent()  # type: ignore
        databricks_code_source = {
            "date": ipython_content["header"]["date"],
            "browserHostName": ipython_content["metadata"]["commandMetadata"]["tags"]["browserHostName"],
            "notebook_path": ipython_content["metadata"]["commandMetadata"]["extraContext"]["notebook_path"],
            "notebook_id": ipython_content["metadata"]["commandMetadata"]["extraContext"]["notebook_id"],
        }
    except Exception as error:
        _logger.debug(f"Code capture databrick failed: {error.__class__.__name__}: {error}. ")
        _logger.info("Extracting the code source version from Databricks failed.")

    return databricks_code_source


def _look_for_git_repository(repo_path: str = ".") -> Repo | None:
    def _log_error(error: str):
        _logger.debug(
            f"Code capture git failed: {error.__class__.__name__}: {error}. "
            "Make sure the current directory is a valid Git repository (non-bare, non worktree) "
            "and its permissions allow the current user to access it."
        )

    try:
        from git import GitError
    except ImportError as error:
        _log_error(str(error))
        return None

    try:
        repo_path = os.path.abspath(repo_path)
    except OSError:
        _logger.debug(f"Code capture failed: the directory '{repo_path}' cannot be accessed by the system")
        return None
    try:
        from git.repo import Repo

        return Repo(repo_path, search_parent_directories=True)
    except GitError as error:
        _log_error(str(error) or repo_path)
        return None


def _extract_databricks_code_version(databricks_code_source: Dict[str, Any]) -> DatabricksVersion:
    url = f"https://{databricks_code_source['browserHostName']}/#notebook/{databricks_code_source['notebook_id']}"
    relative_path = databricks_code_source["notebook_path"]
    date = databricks_code_source["date"]
    databricks_version = DatabricksVersion(url_notebook=url, relative_path=relative_path, timestamp=date)
    return databricks_version


def inform_if_git_repo():
    with hide_logs("vectice"):
        repo = _look_for_git_repository()
    if repo:
        _logger.debug("A git repository was found but code capture is disabled.")


def _extract_git_version(repository: Repo) -> GitVersion | None:
    prefix = "Extracting the Git version failed"
    try:
        url = repository.remotes.origin.url
    except Exception as error:
        _logger.warning(f"{prefix}: we couldn't get the remote URL (no origin?): {error!r}")
        return None
    url = url.rsplit(".git", 1)[0]
    repository_name = url.split("/")[-1]
    try:
        branch_name = repository.active_branch.name
    except Exception as error:
        _logger.warning(f"{prefix}: we couldn't get the branch name (detached mode?): {error!r}")
        return None
    try:
        commit_hash = repository.head.object.hexsha
    except Exception as error:
        _logger.warning(f"{prefix}: we couldn't get the commit hash (no commits?): {error!r})")
        return None
    is_dirty = repository.is_dirty()
    uri = _extract_aws_code_commit(url, repository_name, commit_hash) if "amazonaws" in url else url
    return GitVersion(repository_name, branch_name, commit_hash, is_dirty, uri)


def _extract_aws_code_commit(url: str, repository_name: str, commit_hash: str) -> str:
    region = url.split(".")[1]
    return f"https://{region}.console.aws.amazon.com/codesuite/codecommit/repositories/{repository_name}/commit/{commit_hash}?region={region}"



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict

from vectice.api.json.model_version import ModelVersionStatus
from vectice.api.json.model_version_representation import ModelVersionRepresentationOutput, ModelVersionUpdateInput
from vectice.models.attachment import TAttachment
from vectice.models.attachment_container import AttachmentContainer
from vectice.models.metric import Metric
from vectice.models.property import Property
from vectice.utils.common_utils import (
    convert_list_keyvalue_to_dict,
    format_attachments,
    format_metrics,
    format_properties,
    repr_class,
    strip_dict_list,
)
from vectice.utils.dataframe_utils import repr_list_as_pd_dataframe

_logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from pandas import DataFrame

    from vectice.api.client import Client


class ModelVersionRepresentation:
    """Represents the metadata of a Vectice model version.

    A Model Version Representation shows information about a specific version of a model from the Vectice app.
    It makes it easier to get and read this information through the API.

    NOTE: **Hint**
        A model version ID starts with 'MDV-XXX'. Retrieve the ID in the Vectice App, then use the ID with the following methods to get the model version:
        ```connect.model_version('MDV-XXX')``` or ```connect.browse('MDV-XXX')```
        (see [Connection page](https://api-docs.vectice.com/reference/vectice/connection/#vectice.Connection.model_version)).

    Attributes:
        id (str): The unique identifier of the model version.
        project_id (str): The identifier of the project to which the model version belongs.
        name (str): The name of the model version. For model versions it corresponds to the version number.
        status (str): The status of the model version (EXPERIMENTATION, STAGING, PRODUCTION, or RETIRED).
        description (str): The description of the model version.
        technique (str): The technique used by the model version.
        library (str): The library used by the model version.
        metrics (List[Dict[str, Any]]): The metrics associated with the model version.
        properties (List[Dict[str, Any]]): The properties associated with the model version.
        model_representation (ModelRepresentation): Holds informations about the source model linked to the model version, where all versions are grouped together.
    """

    def __init__(self, output: ModelVersionRepresentationOutput, client: Client):
        from vectice.models.representation.model_representation import ModelRepresentation

        self.id = output.id
        self.project_id = output.project_id
        self.name = output.name
        self.status = output.status
        self.description = output.description
        self.technique = output.technique
        self.library = output.library
        self.metrics = output.metrics
        self.properties = strip_dict_list(output.properties)
        self.model_representation = ModelRepresentation(output.model, client)

        self._client = client
        self._output = output

    def __repr__(self):
        return repr_class(self)

    def asdict(self) -> Dict[str, Any]:
        """Transform the ModelVersionRepresentation into a organised dictionary.

        Returns:
            The object represented as a dictionary
        """
        flat_metrics = convert_list_keyvalue_to_dict(self.metrics)
        flat_properties = convert_list_keyvalue_to_dict(self.properties)

        return {
            "id": self.id,
            "name": self.name,
            "project_id": self.project_id,
            "status": self.status,
            "description": self.description,
            "technique": self.technique,
            "library": self.library,
            "metrics": flat_metrics,
            "properties": flat_properties,
            "model_representation": (self.model_representation.asdict() if self.model_representation else None),
        }

    def metrics_as_dataframe(self) -> DataFrame:
        """Transforms the metrics of the ModelVersionRepresentation into a DataFrame for better readability.

        Returns:
            A pandas DataFrame containing the metrics of the model version.
        """
        return repr_list_as_pd_dataframe(self.metrics)  # change key name

    def properties_as_dataframe(self) -> DataFrame:
        """Transforms the properties of the ModelVersionRepresentation into a DataFrame for better readability.

        Returns:
            A pandas DataFrame containing the properties of the model version.
        """
        return repr_list_as_pd_dataframe(self.properties)  # change key name

    def update(
        self,
        status: str | None = None,
        metrics: dict[str, int | float] | list[Metric] | Metric | None = None,
        properties: dict[str, str | int] | list[Property] | Property | None = None,
        attachments: str | list[str] | None = None,
    ) -> None:
        """Update the Model Version from the API.

        Parameters:
            status: The new status of the model. Accepted values are EXPERIMENTATION, STAGING, PRODUCTION and RETIRED.
            properties: The new properties of the model.
            metrics: The new metrics of the model.
            attachments: The new attachments of the model.

        Returns:
            None
        """
        if status is not None:
            self._update_status(status)

        if attachments is not None:
            self._update_attachments(attachments)

        if metrics is not None:
            self._upsert_metrics(metrics)

        if properties is not None:
            self._upsert_properties(properties)

    def _update_status(self, status: str):
        try:
            status_enum = ModelVersionStatus(status.strip().upper())
        except ValueError as err:
            accepted_statuses = ", ".join([f"{status_enum.value!r}" for status_enum in ModelVersionStatus])
            raise ValueError(f"'{status}' is an invalid value. Please use [{accepted_statuses}].") from err

        model_input = ModelVersionUpdateInput(status=status_enum.value)
        self._client.update_model(self.id, model_input)
        old_status = self.status
        self.status = status_enum.value
        _logger.info(f"Model version {self.id!r} transitioned from {old_status!r} to {self.status!r}.")

    def _upsert_properties(self, properties: dict[str, str | int] | list[Property] | Property):
        clean_properties = list(map(lambda property: property.key_val_dict(), format_properties(properties)))
        new_properties = self._client.upsert_properties("modelVersion", self.id, clean_properties)
        self.properties = strip_dict_list(new_properties)
        _logger.info(f"Model version {self.id!r} properties successfully updated.")

    def _upsert_metrics(self, metrics: dict[str, int | float] | list[Metric] | Metric):
        clean_metrics = list(map(lambda metric: metric.key_val_dict(), format_metrics(metrics)))
        self.metrics = self._client.upsert_metrics("modelVersion", self.id, clean_metrics)
        _logger.info(f"Model version {self.id!r} metrics successfully updated.")

    def _update_attachments(self, attachments: TAttachment):
        container = AttachmentContainer(self._output, self._client)
        container.upsert_attachments(format_attachments(attachments))
        _logger.info(f"Model version {self.id!r} attachments successfully updated.")



from __future__ import annotations

import ast
import logging
import os
import re
import tempfile
from base64 import b64decode
from collections import OrderedDict
from importlib.util import find_spec
from io import BytesIO
from typing import TYPE_CHECKING, Any, TypeVar

import PIL.Image as Image
from typing_extensions import TypedDict

from vectice.api.http_error_handlers import VecticeException
from vectice.autolog.autolog_asset_factory import AssetFactory
from vectice.autolog.model_library import ModelLibrary
from vectice.models.dataset import Dataset
from vectice.models.model import Model
from vectice.models.representation.dataset_representation import DatasetRepresentation
from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
from vectice.models.representation.model_representation import ModelRepresentation
from vectice.models.representation.model_version_representation import ModelVersionRepresentation
from vectice.models.table import Table
from vectice.models.validation import ValidationModel
from vectice.utils.code_parser import FilePathVisitor, VariableVisitor, parse_comments, preprocess_code
from vectice.utils.common_utils import ensure_correct_project_id_from_representation_objs
from vectice.utils.last_assets import _get_asset_parent_name  # pyright: ignore [reportPrivateUsage]

if TYPE_CHECKING:
    from tempfile import TemporaryDirectory

    from pandas import DataFrame
    from pyspark.sql import DataFrame as SparkDF
    from sklearn.pipeline import Pipeline

    # Vectice Object types
    from vectice.autolog.asset_services import TVecticeObjects
    from vectice.autolog.model_types import ModelTypes
    from vectice.models import Phase
    from vectice.models.resource.metadata.db_metadata import TableType

    TModel = TypedDict(
        "TModel",
        {
            "variable": str,
            "model": ModelTypes,
            "library": ModelLibrary,
            "technique": str,
            "metrics": dict,
            "properties": dict,
        },
    )
    DataframeTypes = TypeVar("DataframeTypes", SparkDF, DataFrame)
    TDataset = TypedDict(
        "TDataset",
        {"variable": str, "dataframe": DataframeTypes, "type": TableType},
    )


try:
    from IPython.core.getipython import get_ipython
    from IPython.core.interactiveshell import InteractiveShell
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "To use autolog, please install extra dependencies from vectice package using '%pip install vectice[autolog]'"
    ) from None

is_plotly = False
is_matplotlib = False
if find_spec("plotly") is not None:
    is_plotly = True

if find_spec("matplotlib") is not None:
    is_matplotlib = True

_logger = logging.getLogger(__name__)

NOTEBOOK_CELLS = {}
GRAPHS = []
running = False


def _get_cell_id_and_plot_name(ipython: Any) -> tuple[str, str]:
    # Collab
    # cell_id = IP.get_parent()["metadata"]["colab"]["cell_id"]

    # Databricks
    # cell_id = IP.get_parent()["metadata"]["commandId"]
    cell_metadata = ipython.get_parent()["metadata"]  # type: ignore
    if cell_metadata.get("colab") and "cell_id" in cell_metadata.get("colab"):  # type: ignore
        cell_id = cell_metadata["colab"]["cell_id"]
        plot_name = f"vect_plot_{cell_id}"
    elif cell_metadata.get("commandId"):  # type: ignore
        cell_id = cell_metadata["commandId"]
        plot_name = f"vect_plot_{cell_id[-6:]}"
    else:
        cell_id = cell_metadata["cellId"]
        plot_name = (
            f"vect_plot_{cell_id.split('-')[-1]}"
            if "vscode-notebook-cell" not in cell_id
            else f"vect_plot_{cell_id.split('ipynb#')[-1]}"
        )
    return cell_id, plot_name


def start_listen():
    ipython = get_ipython()  # type: ignore

    def _notebook_cell_content_update():
        global NOTEBOOK_CELLS

        cell_id, _ = _get_cell_id_and_plot_name(ipython)
        cell_content = ipython.get_parent()["content"]["code"]  # type: ignore
        NOTEBOOK_CELLS[cell_id] = cell_content

    def _check_graph_libraries(cell_content: str) -> bool:
        # check for imports and recommended aliases
        if "matplotlib" in cell_content or "plt" in cell_content:
            return True
        if "seaborn" in cell_content or "sns" in cell_content:
            return True
        if "plotly" in cell_content or "px" in cell_content:
            _logger.warning("Plotly is not supported currently.")
            return False
        return False

    def _get_graphs(remove: bool) -> list[str]:

        # check for a valid matplotlib extension, this covers seaborn aswell
        def is_valid_matplotlib_extension(filename: str) -> bool:
            valid_extensions = {".png", ".pdf", ".ps", ".eps", ".svg", ".jpg", ".jpeg", ".tiff", ".tif", ".bmp"}
            _, extension = os.path.splitext(filename)
            return extension.lower() in valid_extensions

        cell_content = ipython.get_parent()["content"]["code"]  # type: ignore
        # get commented lines and not lines with comments
        lines = cell_content.split("\n")
        matched_lines = "\n".join(
            [line for line in lines if (remove and re.search(r"^#.*", line)) or not re.search(r"^#.*", line)]
        )
        cell = preprocess_code(matched_lines)
        tree = ast.parse(cell)
        visitor = FilePathVisitor(is_graph_path=True)
        visitor.visit(tree)

        graph_paths = []
        for graph in visitor.file_paths:
            if is_valid_matplotlib_extension(graph) and "http" not in graph:
                graph_paths.append(graph)
        return graph_paths

    def _remove_commented_graphs():
        # removes a saved graph if commented out
        graphs = _get_graphs(True)

        for graph in graphs:
            if graph in GRAPHS:
                GRAPHS.remove(graph)

    def _get_saved_graphs() -> list[str]:
        graphs = _get_graphs(False)

        for graph in graphs:
            if graph not in GRAPHS:
                GRAPHS.append(graph)
        return graphs

    def _capture_graph():
        # globals allows the function to get the variable, or it outside it's scope
        global GRAPHS
        # running variable prevents recursion
        global running

        cell_content = ipython.get_parent()["content"]["code"]  # type: ignore
        # we generate plot names that are then used in the cell magic
        _, plot_name = _get_cell_id_and_plot_name(ipython)
        # check if graphs have been commented out before getting saved graphs
        _remove_commented_graphs()
        # get the saved graphs
        saved_graphs = _get_saved_graphs()
        # check if running already and that there aren't saved graphs as saved graphs are the priority
        if not running and not saved_graphs and _check_graph_libraries(cell_content):
            running = True
            # Remove autolog from cell so we don't create an extra iteration
            clean_cell = re.sub(r"autolog.*?$", "", cell_content)
            ipython.run_cell_magic("capture", plot_name, clean_cell)  # type: ignore
            # Check if plot name is already in our list as we don't want duplicates
            if plot_name not in GRAPHS:
                GRAPHS.append(plot_name)
        # remove event graph if save graph true, prevents double capturing
        if saved_graphs and plot_name in GRAPHS:
            GRAPHS.remove(plot_name)
        running = False

    def _load_ipython_extension():
        """Load the extension in IPython."""
        ipython.events.register("pre_execute", _capture_graph)  # type: ignore
        ipython.events.register("post_execute", _notebook_cell_content_update)  # type: ignore

    if ipython and not any([ev for ev in ipython.events.callbacks["post_execute"] if "_notebook_cell_content_update" in str(ev)]):  # type: ignore
        _load_ipython_extension()


####### Autolog logic
class Autolog:
    def __init__(
        self,
        phase: Phase | None,
        ipy: InteractiveShell,
        is_notebook: bool,
        create: bool = True,
        note: str | None = None,
        capture_schema_only: bool = True,
        capture_comments: bool = True,
    ):
        from vectice.services.phase_service import PhaseService

        if phase is None:
            raise ValueError("Login")

        if create is True:
            iteration = phase.create_iteration()
        else:
            iteration = PhaseService(
                phase._client  # pyright: ignore[reportPrivateUsage]
            ).get_active_iteration_or_create(phase)

        self._iteration = iteration
        self._iteration_service = iteration._service  # pyright: ignore[reportPrivateUsage]
        self._capture_schema_only = capture_schema_only
        self._ip = ipy
        self._local_vars = self._ip.user_global_ns
        self._capture_comments = capture_comments
        self._is_notebook = is_notebook
        self._cell_content = self._get_notebook_cell_content() if is_notebook is True else self._get_cell_content()
        self._vectice_data = self._get_variable_matches(self._local_vars)
        self._failed_assets = []

        # Get back objects to log
        self._assets = self._get_assets()
        graphs = self._get_graphs(is_notebook)

        if note:
            self._iteration_service.log_comment(note)
            _logger.info(f"Note logged in iteration {self._iteration.name!r}.")

        # Log objects

        self._log_assets(self._assets)

        if graphs:
            for graph in graphs:
                try:
                    self._iteration_service.log_image_or_file(graph)
                    graph = graph if isinstance(graph, str) else graph.filename
                    _logger.info(f"Graph {graph!r} logged in iteration {self._iteration.name!r}.")
                except Exception as e:
                    self._failed_assets.append({"reason": e, "asset": graph, "type": "Graph"})

        if len(self._failed_assets):
            _logger.warning("The following assets failed to log:")
        for failed_asset in self._failed_assets:
            _logger.warning(f"{failed_asset['type']} {failed_asset['asset']!r}, reason: {failed_asset['reason']}")

    def _get_variable_matches(self, local_vars: dict[str, Any]) -> list[dict[Any, Any]]:
        vectice_data = []
        all_cell_vars = set()
        all_vectice_calls = set()
        for cell_not_processed in self._cell_content:
            cell = preprocess_code(cell_not_processed)
            variable_comments = parse_comments(cell_not_processed) if self._capture_comments else []
            vectice_match = {}
            all_vars: OrderedDict[str, Any] = OrderedDict()

            vectice_match["cell"] = cell
            vectice_match["comments"] = variable_comments

            # Parse the cell content using the VariableVisitor
            try:
                tree = ast.parse(cell)
            except SyntaxError as err:
                raise SyntaxError(
                    "Autolog is unable to parse the code. Make sure all non-Python syntax, such as bash commands, are properly indented and preceded by '%' or '!'. If the error persist, please contact your sales representative."
                ) from err

            visitor = VariableVisitor()
            visitor.visit(tree)
            # Keep set of all vectice call vars
            all_vectice_calls = all_vectice_calls.union(visitor.vectice_call_vars)
            # Update all_vars with variable names and values in the order they appear
            for var in visitor.variables:
                if var in local_vars and var not in all_vars:
                    all_vars[var] = local_vars[var]
                # check if the var exists in a previous cell for autolog.notebook before the vectice object uses the var
                if self._is_notebook:
                    all_vectice_calls = {vect_var for vect_var in all_vectice_calls if vect_var not in all_cell_vars}
                # Keep track of all vars
                all_cell_vars.add(var)

            vectice_match["variables"] = all_vars
            vectice_data.append(vectice_match)
        clean_vectice_data = []
        for data in vectice_data:
            for var in all_vectice_calls:
                # the order of cells misses vars
                if var in data["variables"]:
                    del data["variables"][var]
            # check that there are vars for the cell and then append or if there are comments
            if len(data["variables"]) >= 1 or len(data["comments"]) >= 1:
                clean_vectice_data.append(data)
        return clean_vectice_data

    def _get_notebook_cell_content(self) -> list[Any]:
        cell_id, _ = _get_cell_id_and_plot_name(self._ip)
        cell_content = self._ip.get_parent()["content"]["code"]  # type: ignore
        NOTEBOOK_CELLS[cell_id] = cell_content
        return list(NOTEBOOK_CELLS.values())

    def _get_cell_content(self) -> list[Any]:
        """Used by autolog cell to get the content of the cell. This is used to parse for variables."""
        cell_content = self._ip.get_parent()["content"]["code"]  # pyright: ignore[reportAttributeAccessIssue]
        if cell_content is None:
            raise ValueError("Failed to get cell content.")
        return [cell_content]

    def _get_asset_comments(self, comments: dict, key: str) -> tuple[str | None, str | None]:
        try:
            comment_before = comments["comment_matches_before"][key]["comment"]
        except KeyError:
            comment_before = None
        try:
            comment_after = comments["comment_matches_after"][key]["comment"]
        except KeyError:
            comment_after = None
        return comment_before, comment_after

    def _get_assets(self) -> list[TModel | TDataset | TVecticeObjects]:
        assets = []
        for data in self._vectice_data:
            for key in data["variables"].keys():
                # skip cell inputs/outputs
                if key.startswith("_"):
                    continue
                asset = data["variables"][key]
                try:
                    asset_information = AssetFactory.get_asset_service(key, asset, data).get_asset()
                    if asset_information is not None:
                        assets.append(asset_information)
                except VecticeException:
                    pass
        return self._deduplicate_assets(assets)

    def _deduplicate_assets(self, assets: list[dict[str, Any]]) -> list[TModel | TDataset | TVecticeObjects]:
        unique_dict = {}
        result = []

        for item in assets:
            variable = item["variable"]

            if variable not in unique_dict.keys():
                # If variable is not seen before, add it to the dictionary
                unique_dict[variable] = item
                result.append(item)
            else:
                existing_item = unique_dict[variable]
                if "vectice_object" in item:
                    existing_item["vectice_object"] = item["vectice_object"]
                elif "dataframe" in item:
                    existing_item["type"] = item["type"]
                else:
                    existing_item["library"] = item["library"]

                    # Merge dictionaries for the "metrics" attribute
                    existing_item_metrics = existing_item["metrics"]
                    item_metrics = item["metrics"]

                    for key, value in item_metrics.items():
                        existing_item_metrics[key] = value

        return result

    def _get_pipeline_steps(self, pipeline: Pipeline) -> dict:
        sklearn_pipeline = {}
        try:
            for step in pipeline.steps:
                step_name, step_obj = step
                if hasattr(step_obj, "named_transformers_"):
                    for k, v in step_obj.named_transformers_.items():
                        sklearn_pipeline[f"pipeline_{step_name}_{k}_steps"] = list(v.named_steps.keys())
            return sklearn_pipeline
        except Exception:
            return sklearn_pipeline

    def _get_pipeline_info(self, pipeline: Pipeline) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        # if pipeline with classifier
        try:
            from sklearn.base import is_classifier, is_regressor

            model = pipeline.steps[-1][-1]
            model_name = pipeline.steps[-1][0]
            if is_regressor(model) or is_classifier(model):
                model_params = {
                    f"{model_name}_{key}": value
                    for key, value in model.get_params().items()
                    if value is not None and bool(str(value))
                }
                model_params.update(self._get_pipeline_steps(pipeline))
                return "sklearn-pipeline", model_params
        except Exception:
            pass

        try:
            pipeline_params = {
                str(key): value
                for key, value in pipeline.get_params().items()
                if value is not None and bool(str(value))
            }
            return "sklearn-pipeline", pipeline_params
        except Exception:
            return None, None

    def _get_all_cells_comments(self) -> list[dict]:
        # Get all comments from all cells
        all_comments_and_variables = []
        for data in self._vectice_data:
            if data["comments"]:
                for comments in data["comments"]:
                    all_comments_and_variables.append(comments)
        return all_comments_and_variables

    def _get_comments_before(
        self, all_comments_and_variables: list[dict], asset: TModel | TDataset | TVecticeObjects, comments_logged: list
    ):
        comments_to_log = []
        for data in all_comments_and_variables:
            # If we find the variable we log all the comments found
            if data["variable"] == asset["variable"]:
                return comments_to_log
            # We only want comments remaining
            if data["comment"] not in comments_logged:
                comments_to_log.append(data["comment"])
        # return nothing if we don't see the asset
        return []

    def _log_comments_before_asset(
        self, all_comments_and_variables: list[dict], asset: TModel | TDataset | TVecticeObjects, comments_logged: list
    ) -> list:
        # Get comments to log before each asset
        comments_before = self._get_comments_before(all_comments_and_variables, asset, comments_logged)
        # Log comments before each asset
        if comments_before:
            for comment in comments_before:
                if comment:
                    self._iteration_service.log_comment(comment)
            # Keep track of what's logged
            return comments_logged + comments_before
        return comments_logged

    def _log_remaining_comments(self, all_unique_comments: set, comments_logged: list) -> None:
        # Ensure last asset comments are filtered out
        all_comments_logged = set(comments_logged)
        still_comments_to_log = list(all_unique_comments - all_comments_logged)
        # Log remaining comments after last asset
        if still_comments_to_log:
            for comment in still_comments_to_log:
                if comment:
                    self._iteration_service.log_comment(comment)
        # Logging for comments captured
        all_comments_logged.discard(None)
        filtered_comments = list(filter(lambda comment: comment is not None, still_comments_to_log))
        if filtered_comments or all_comments_logged:
            _logger.info(f"Comments logged in iteration {self._iteration.name!r}.")

    def _log_assets(
        self,
        assets: list[TModel | TDataset | TVecticeObjects],
    ):
        from vectice.services import iteration_service

        # all comments and variables
        all_comments_and_variables = self._get_all_cells_comments()
        # Unique comment set
        all_unique_comments = set([data["comment"] for data in all_comments_and_variables])
        # Keep track of what is logged
        comments_logged = []
        for asset in assets:
            try:
                comments_logged = self._log_comments_before_asset(all_comments_and_variables, asset, comments_logged)
                if "vectice_object" in asset:
                    self._log_vectice_asset(asset)
                elif "model" in asset:
                    self._log_model(asset)
                elif "dataframe" in asset:
                    self._log_dataset(asset)
            except Exception as e:
                if "vectice_object" in asset:
                    asset_type = "Vectice object"
                elif "model" in asset:
                    asset_type = "Model"
                else:
                    asset_type = "Dataset"
                self._failed_assets.append({"reason": e, "asset": asset["variable"], "type": asset_type})
        # Log the remaining comments
        self._log_remaining_comments(all_unique_comments, comments_logged)
        # After logging, don't re-use code file
        iteration_service.lineage_file_id = None

    def _check_for_single_model(self):
        model_list = [asset for asset in self._assets if "model" in asset]
        if len(model_list) > 1:
            return False
        return True

    def _get_single_model_metrics(self) -> dict:
        from vectice.autolog.asset_services.metric_service import MetricService

        # get a single models metrics for notebook function
        metrics = {}
        if not self._check_for_single_model():
            return {}
        for data in self._vectice_data:
            cell_metrics = MetricService(data["cell"])._get_model_metrics(data)  # type: ignore[reportPrivateUsage]
            if cell_metrics:
                metrics.update(cell_metrics)
        return metrics

    def _log_model(self, model: TModel):
        temp_dir: TemporaryDirectory | None = None
        temp_file_path: str | None = None
        if model["library"] is ModelLibrary.KERAS:
            graph = None

            try:
                from keras.utils import plot_model  # type: ignore[reportMissingImports]

                temp_dir = tempfile.TemporaryDirectory()
                file_name = f"{model['variable']!s}_plot.png"
                temp_file_path = rf"{temp_dir.name}\{file_name}"
                graph = plot_model(model["model"], to_file=temp_file_path, show_shapes=True, show_layer_names=False)

                if graph is None:
                    temp_file_path = None
                    _logger.info(
                        "Unable to generate the model plot. Please check the 'plot_model' function from the Keras library is working correctly. Make sure that the graphviz and pydot packages are installed and configured properly."
                    )

            except Exception as e:
                _logger.info(
                    f"Unable to generate the model plot. Ensure that the Keras library is correctly installed and up-to-date. Error details: {e}"
                )

        model_object = model["model"]

        library, algorithm, params = model["library"], model["technique"], model["properties"]
        if library is ModelLibrary.STATSMODEL:
            model_summary = model_object.summary().as_text()  # type: ignore[reportAttributeAccessIssue]
        else:
            model_summary = None

        model_name = f"{self._iteration.phase.id}-{model['variable']}"
        # safety check, if no metrics and single model, get metrics
        model_metrics = model["metrics"] if model["metrics"] else self._get_single_model_metrics()
        temp_files = []
        if library is ModelLibrary.SKLEARN_PIPELINE:
            temp_dir, temp_json_file_path, temp_html_file_path = self._get_sklearn_pipeline(model_object, model_name)
            if temp_json_file_path:
                temp_files.append(temp_json_file_path)
            if temp_html_file_path:
                temp_files.append(temp_html_file_path)

        if library is ModelLibrary.SKLEARN_SEARCH:
            tables = self._get_sklearn_search_results_and_space_tables(model_object, model_name)
            temp_files.extend(tables)

        self._iteration_service.log_model(
            Model(
                library=library.value,
                technique=algorithm,
                metrics=model_metrics,
                properties=params,
                name=model_name,
                predictor=model_object,
                attachments=temp_file_path or temp_files,
            )
        )
        _logger.info(f"Model {model_name!r} logged in iteration {self._iteration.name!r}.")

        if model_summary:
            # log statsmodel summary
            self._iteration_service.log_comment(model_summary)

        if temp_dir is not None:
            temp_dir.cleanup()

    def _get_sklearn_pipeline(
        self, pipeline: ModelTypes, model_name: str
    ) -> tuple[TemporaryDirectory[str], str, str] | tuple[None, None, None]:
        temp_dir = None
        try:
            from sklearn.utils import estimator_html_repr

            from vectice.utils.sklearn_pipe_utils import pipeline_to_json

            temp_dir = tempfile.TemporaryDirectory()
            json_file_name = f"{model_name!s}_pipeline.json"
            html_file_name = f"{model_name!s}_pipeline.html"
            temp_json_file_path = rf"{temp_dir.name}\{json_file_name}"
            temp_html_file_path = rf"{temp_dir.name}\{html_file_name}"

            pipeline_json = pipeline_to_json(pipeline)
            pipeline_html = estimator_html_repr(pipeline)
            if pipeline_json:
                with open(temp_json_file_path, "w") as json_file:
                    json_file.write(pipeline_json)

            if pipeline_html:
                with open(temp_html_file_path, "w", encoding="utf-8") as html_file:
                    html_file.write(pipeline_html)

            return temp_dir, temp_json_file_path, temp_html_file_path
        except Exception:
            return None, None, None

    def _get_sklearn_search_results_and_space_tables(self, model: ModelTypes, model_name: str) -> list:
        tables = []
        try:
            import pandas as pd

            results_df = pd.DataFrame(model.cv_results_)  # pyright: ignore[reportAttributeAccessIssue]
        except Exception:
            return tables
        try:
            tables.append(self._get_sklearn_search_space(model, model_name, results_df))
        except Exception:
            pass
        try:
            tables.append(self._get_sklearn_search_results(model, model_name, results_df))
        except Exception:
            pass
        return tables

    def _get_sklearn_search_results(self, model: ModelTypes, model_name: str, results_df: DataFrame) -> Table:
        sorted_df = results_df.sort_values(by="rank_test_score")
        top_scores_df = sorted_df.head(5)
        return Table(top_scores_df, name=f"{model_name}_search_results")

    def _get_sklearn_search_space(self, model: ModelTypes, model_name: str, results_df: DataFrame) -> Table:
        import pandas as pd

        param_columns = [col for col in results_df.columns if "param" in col and "params" not in col]
        data_dict = {}
        for param in param_columns:
            try:
                data_dict[f"{param} (min,max)"] = [[results_df[param].min(), results_df[param].max()]]
            except Exception:
                data_dict[f"{param} (uniques)"] = [results_df[param].unique()]

        params_df = pd.DataFrame(dict([(k, pd.Series(v)) for k, v in data_dict.items()]))
        # Transpose the DataFrame
        df_transposed = params_df.transpose()
        # Reset the index to move row headers into a column
        df_transposed_reset = df_transposed.reset_index()
        # Rename the index column for clarity
        df_transposed_reset = df_transposed_reset.rename(columns={"index": "Parameters", 0: "Values"})
        return Table(df_transposed_reset, name=f"{model_name}_search_space")

    def _log_vectice_asset(self, asset: TVecticeObjects) -> None:

        asset_to_log = asset["vectice_object"]
        if isinstance(asset_to_log, ValidationModel):
            object_name = asset["variable"]
        else:
            object_name = asset_to_log.name

        ensure_correct_project_id_from_representation_objs(self._iteration.project.id, asset_to_log)

        if isinstance(asset_to_log, Dataset):
            self._iteration_service.log_dataset(asset_to_log)
            _logger.info(f"Dataset {object_name!r} logged in iteration {self._iteration.name!r}.")
        elif isinstance(asset_to_log, Table):
            self._iteration_service.log_table(asset_to_log)
            _logger.info(f"Table {object_name!r} logged in iteration {self._iteration.index!r}.")
        elif isinstance(asset_to_log, Model):
            if asset_to_log.predictor:
                self._iteration._assign_model_predictor_metadata(asset_to_log)  # pyright: ignore[reportPrivateUsage]
            self._iteration_service.log_model(asset_to_log)
            _logger.info(f"Model {object_name!r} logged in iteration {self._iteration.index!r}.")
        elif isinstance(
            asset_to_log,
            (ModelRepresentation, ModelVersionRepresentation, DatasetRepresentation, DatasetVersionRepresentation),
        ):
            asset_type, already_assigned = self._iteration_service.assign_version_representation(
                asset_to_log
            )  # pyright: ignore[reportPrivateUsage]
            link = "already linked" if already_assigned else "linked"
            asset_name_variable = asset["variable"]
            is_parent = isinstance(asset_to_log, (ModelRepresentation, DatasetRepresentation))
            asset_version = (
                asset_to_log._last_version.name  # pyright: ignore [reportPrivateUsage]
                if (is_parent and asset_to_log._last_version is not None)  # pyright: ignore [reportPrivateUsage]
                else asset_to_log.name
            )
            asset_name = (
                asset_to_log.name
                if is_parent
                else _get_asset_parent_name(asset_to_log)  # pyright: ignore [reportPrivateUsage]
            )
            _logger.info(
                f"{asset_type} {asset_name!r} {asset_version} named {asset_name_variable!r} as a variable {link} to iteration {self._iteration.index!r}."
            )
        else:
            self._iteration._log_validation_model(asset_to_log)  # pyright: ignore[reportPrivateUsage]

    def _log_dataset(self, dataset: TDataset) -> None:
        from vectice import Dataset, DatasetType, FileResource, NoResource

        resource = self._get_dataset_resource(dataset)
        dataset_name = f"{self._iteration.phase.id}-{dataset['variable']}"
        no_resource_dataset = Dataset(
            type=DatasetType.UNKNOWN,
            resource=NoResource(
                dataframes=dataset["dataframe"],
                origin="DATAFRAME",
                type=dataset["type"],
                capture_schema_only=self._capture_schema_only,
            ),
            name=dataset_name,
        )
        if resource:
            # TODO Dataset type ?
            vec_dataset = Dataset(
                type=DatasetType.UNKNOWN,
                resource=FileResource(
                    paths=resource,
                    dataframes=dataset["dataframe"],
                    capture_schema_only=self._capture_schema_only,
                ),
                name=dataset_name,
            )
        else:
            vec_dataset = no_resource_dataset
        try:
            self._iteration_service.log_dataset(vec_dataset)
        except FileNotFoundError:
            self._iteration_service.log_dataset(no_resource_dataset)
        _logger.info(f"Dataset {dataset_name!r} logged in iteration {self._iteration.name!r}.")

    def _get_dataset_resource(self, dataset: TDataset) -> str | None:
        import re

        if not self._cell_content:
            return None
        try:
            # Avoid getting stray dataset with autolog.notebook()

            match = []
            for cell in self._cell_content:
                cell = preprocess_code(cell)
                tree = ast.parse(cell)
                visitor = FilePathVisitor(is_dataset_path=True)
                visitor.visit(tree)
                for resource_info in visitor.dataset_file_paths:
                    path_var, file_path = resource_info["variable"], resource_info["path"]
                    dataset_variable = re.escape(dataset["variable"])
                    if path_var:
                        pattern = rf"{dataset_variable}\s*=\s*pd\.read_csv\(\s*{path_var}\s*\).*\n"
                    else:
                        regex_path = re.escape(file_path)
                        pattern = rf"{dataset_variable}\s*=\s*pd\.read_csv\(.*?{regex_path}.*?\).*\n"
                    match_dataset_path = re.search(pattern, cell)
                    if match_dataset_path:
                        match.append(file_path)

            if len(match) < 1:
                return None
            # TODO update regex
            # check if read csv has comma dominated arguments
            return match[0]
        except TypeError:
            return None

    def _get_image_data(self, graph: str) -> str | None:
        try:
            captured = self._local_vars[graph]
        except KeyError:
            return None
        try:
            return captured.outputs[0].data["image/png"]
        except IndexError:
            return None

    def _get_graphs(self, is_notebook: bool) -> list[Any]:
        global GRAPHS

        _, plot_name = _get_cell_id_and_plot_name(self._ip)

        graphs = []
        # Create png data and don't convert saved graphs
        graphs_data = [
            (graph, self._get_image_data(graph))
            for graph in GRAPHS
            if not os.path.isfile(graph) and self._get_image_data(graph)
        ]
        # Create the Images from the png_data from the cell magic
        for graph_name, png_bytes in graphs_data:
            if isinstance(png_bytes, str):
                png_bytes = b64decode(png_bytes)
                bytes_io = BytesIO(png_bytes)
                image = Image.open(bytes_io)
                image.filename = f"{graph_name}.png"  # type: ignore
                # Check if it is cell event graph and make sure graph belongs to cell with the plot_name /  if notebook get all graph png data. Else statement causes duplicates
                if (not is_notebook and plot_name == graph_name) or is_notebook:
                    graphs.append(image)

        for graph in GRAPHS:
            is_str_graph = isinstance(graph, str)
            is_saved_graph_in_cell = "savefig" in str(self._cell_content) or "write_image" in str(self._cell_content)
            is_cell_condition = not is_notebook and graph in str(self._cell_content) and is_saved_graph_in_cell
            is_notebook_condition = is_notebook and os.path.isfile(graph)
            # check for cell saved graphs, check if graph is in cell as multiple cells can be run / check for notebook saved graphs, filepath check makes sure it's a saved graph and not a plot_name
            if is_str_graph and (is_cell_condition or is_notebook_condition):
                graphs.append(graph)
        return graphs



from vectice.autolog.asset_services.catboost_service import AutologCatboostService
from vectice.autolog.asset_services.keras_service import AutologKerasService
from vectice.autolog.asset_services.lightgbm_service import AutologLightgbmService
from vectice.autolog.asset_services.pandas_service import AutologPandasService
from vectice.autolog.asset_services.pyspark_service import AutologPysparkService
from vectice.autolog.asset_services.pytorch_service import AutologPytorchService
from vectice.autolog.asset_services.sklearn_service import AutologSklearnService
from vectice.autolog.asset_services.statsmodel_wrapper_service import AutologStatsModelWrapperService
from vectice.autolog.asset_services.vectice_asset_service import (
    AutologVecticeAssetService,
    TVecticeObjects,
    VecticeObjectClasses,
    VecticeObjectTypes,
)

__all__ = [
    "AutologPandasService",
    "AutologPysparkService",
    "AutologPytorchService",
    "AutologCatboostService",
    "AutologKerasService",
    "AutologLightgbmService",
    "AutologSklearnService",
    "AutologStatsModelWrapperService",
    "AutologVecticeAssetService",
    "TVecticeObjects",
    "VecticeObjectTypes",
    "VecticeObjectClasses",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from vectice.autolog.asset_services.metric_service import MetricService
from vectice.autolog.asset_services.technique_service import TechniqueService
from vectice.autolog.model_library import ModelLibrary

if TYPE_CHECKING:
    from catboost.core import CatBoost


class AutologCatboostService(MetricService, TechniqueService):
    def __init__(self, key: str, asset: CatBoost, data: dict):
        self._asset = asset
        self._key = key

        super().__init__(cell_data=data)

    def get_asset(self):
        _, params = self._get_catboost_info(self._asset)
        return {
            "variable": self._key,
            "model": self._asset,
            "library": ModelLibrary.CATBOOST,
            "technique": self._get_model_technique(self._asset, ModelLibrary.CATBOOST),
            "metrics": self._get_model_metrics(self._cell_data),
            "properties": params,
        }

    def _get_catboost_info(self, model: CatBoost) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        try:
            params = {str(key): value for key, value in model.get_all_params().items() if value is not None}
            return "catboost", params
        except AttributeError:
            return None, None



from __future__ import annotations

from enum import Enum
from typing import Any

from vectice.api.json.json_type import TJSON
from vectice.api.json.project import ProjectOutput


class PhaseStatus(Enum):
    """Enumeration of the different phase statuses."""

    NotStarted = "NotStarted"
    InProgress = "Draft"
    Completed = "Completed"
    InReview = "InReview"


class PhaseOutput(TJSON):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._project: ProjectOutput | None = None
        if "parent" in self:
            self._project = ProjectOutput(**self["parent"])

    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def index(self) -> int:
        return int(self["index"])

    @property
    def status(self) -> PhaseStatus:
        return PhaseStatus(self["status"])

    @property
    def project(self) -> ProjectOutput | None:
        return self._project

    @property
    def steps_count(self) -> int | None:
        return int(self["stepsCount"]) if "stepsCount" in self else None

    @property
    def iterations_count(self) -> int | None:
        return (
            int(self["iterationsCount"]["total"])
            if "iterationsCount" in self and "total" in self["iterationsCount"]
            else None
        )

    @property
    def active_iterations_count(self) -> int | None:
        if "iterationsCount" in self:
            iter_count = self["iterationsCount"]
            not_started = int(iter_count["notStarted"])
            in_progress = int(iter_count["inProgress"])
            in_review = int(iter_count["inReview"])
            return not_started + in_progress + in_review

        return None



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone


@dataclass
class Property:
    """Defines a property for a model or a dataset.

    Parameters:
        key: The key to identify the property.
        value: The value of the property.
        timestamp: The timestamp of the property. Corresponds to the
            property creation time if not explicitly passed.
        name: The name of the property.
    """

    key: str
    value: str
    timestamp: datetime | str = datetime.now(timezone.utc).isoformat()
    name: str | None = None

    def __post_init__(self):
        if not isinstance(self.value, str):  # pyright: ignore[reportUnnecessaryIsInstance]
            self.value = str(self.value)

    def __repr__(self) -> str:
        return f"Property(key='{self.key}', value='{self.value}')"

    def key_val_dict(self) -> dict[str, str | int]:
        return {"key": self.key, "value": self.value}



from __future__ import annotations

from vectice.models.resource.base import Resource
from vectice.models.resource.bigquery_resource import BigQueryResource
from vectice.models.resource.databricks_table_resource import DatabricksTableResource
from vectice.models.resource.file_resource import File, FileResource, FilesMetadata
from vectice.models.resource.gcs_resource import GCSResource, NoSuchGCSResourceError
from vectice.models.resource.no_resource import NoResource
from vectice.models.resource.s3_resource import NoSuchS3ResourceError, S3Resource
from vectice.models.resource.snowflake_resource import SnowflakeResource
from vectice.models.resource.spark_table_resource import SparkTableResource

__all__ = [
    "Resource",
    "DatabricksTableResource",
    "FilesMetadata",
    "FileResource",
    "File",
    "GCSResource",
    "BigQueryResource",
    "NoSuchGCSResourceError",
    "NoResource",
    "S3Resource",
    "NoSuchS3ResourceError",
    "SnowflakeResource",
    "SparkTableResource",
]



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, cast

from vectice.api._utils import read_nodejs_date
from vectice.api.json.artifact_version import ArtifactVersion
from vectice.api.json.json_type import TJSON
from vectice.api.json.model import ModelOutput

if TYPE_CHECKING:
    from datetime import datetime


class ModelVersionStatus(Enum):
    """Enumeration of the different model statuses."""

    EXPERIMENTATION = "EXPERIMENTATION"
    STAGING = "STAGING"
    PRODUCTION = "PRODUCTION"
    RETIRED = "RETIRED"


class ModelVersionInput(TJSON):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str:
        return str(self["description"])

    @property
    def repository(self) -> str:
        return str(self["repository"])

    @property
    def algorithm_name(self) -> str:
        return str(self["algorithmName"])

    @property
    def status(self) -> ModelVersionStatus:
        return ModelVersionStatus[str(self["status"])]

    @property
    def framework(self) -> str:
        return str(self["framework"])

    @property
    def uri(self) -> str:
        return str(self["uri"])

    @property
    def is_starred(self) -> str:
        return str(self["isStarred"])

    @property
    def version(self) -> int:
        return int(self["version"])


class ModelVersionOutput(TJSON):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        if "model" in self:
            self._model: ModelOutput = ModelOutput(**self["model"])
        else:
            self._model = None  # type: ignore

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str | None:
        if self.get("description", None):
            return str(self["description"])
        else:
            return None

    @property
    def created_date(self) -> datetime | None:
        return read_nodejs_date(self["createdDate"])

    @property
    def updated_date(self) -> datetime | None:
        if "updatedDate" in self:
            return read_nodejs_date(self["updatedDate"])
        else:
            return None

    @property
    def version(self) -> ArtifactVersion:
        return ArtifactVersion(version_number=self.version_number, version_name=self.name, version_id=self.id)

    @property
    def version_number(self) -> int:
        return int(self["versionNumber"])

    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def author_id(self) -> int:
        return int(self["authorId"])

    @property
    def deleted_date(self) -> datetime | None:
        if "deletedDate" in self:
            return read_nodejs_date(self["deletedDate"])
        else:
            return None

    @property
    def project(self) -> str:
        return str(self["project"])

    @property
    def algorithm_name(self) -> str:
        return str(self["algorithmName"])

    @property
    def status(self) -> ModelVersionStatus:
        return ModelVersionStatus[str(self["status"])]

    @property
    def resources(self):
        return self.get("resources", None)

    @property
    def framework(self) -> str:
        return str(self["framework"])

    @property
    def uri(self) -> str:
        return str(self["uri"])

    @property
    def is_starred(self) -> bool:
        return bool(self["isStarred"])

    @property
    def model_id(self) -> int:
        return int(self["modelId"])

    @property
    def model(self) -> ModelOutput:
        return self._model

    @property
    def metrics(self) -> dict[str, Any]:
        if "metrics" in self and self["metrics"] is not None:
            return cast(dict, self["metrics"])
        else:
            return {}

    @property
    def lineage(self) -> dict:
        return self["origin"]

    @property
    def origin_id(self) -> int:
        return int(self["origin"]["id"])



from __future__ import annotations

import textwrap


class HttpError(Exception):
    def __init__(self, code: int, reason: str, path: str, method: str, json: str | None):
        super().__init__()
        self.code: int = code
        self.reason: str = reason
        self.path = path
        self.method = method
        self.json = json

    def __str__(self):
        begin = textwrap.dedent(
            f"""
            HTTP Error Code {self.code} : {self.reason}
            {self.method} {self.path}
        """
        )
        if self.json:
            if "apiKey" in self.json:
                end = ""
            else:
                end = textwrap.dedent(
                    f"""
                    ---- payload ---
                    {self.json}
                    ---
                    """
                )
        else:
            end = ""
        return begin + end



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from vectice.autolog.asset_services.metric_service import MetricService
from vectice.autolog.asset_services.technique_service import TechniqueService
from vectice.autolog.model_library import ModelLibrary

if TYPE_CHECKING:
    from keras import Model as KerasModel  # type: ignore[reportMissingImports]  # type: ignore[reportMissingImports]
    from keras.layers import InputLayer  # type: ignore[reportMissingImports]
    from keras.models import Model as KerasModel  # type: ignore[reportMissingImports]


class AutologKerasService(MetricService, TechniqueService):
    def __init__(self, key: str, asset: KerasModel, data: dict):
        self._asset = asset
        self._key = key

        super().__init__(cell_data=data)

    def get_asset(self):
        model_metrics = self._get_model_metrics(self._cell_data)
        training_metrics = self._get_keras_training_metrics(self._asset)
        _, params = self._get_keras_info(self._asset)
        return {
            "variable": self._key,
            "model": self._asset,
            "library": ModelLibrary.KERAS,
            "technique": self._get_model_technique(self._asset, ModelLibrary.KERAS),
            "metrics": {**model_metrics, **training_metrics},
            "properties": params,
        }

    def _format_keras_params(self, model: KerasModel) -> dict[str, Any]:
        params: dict[str, Any] = {}

        def _get_output_shape(layer: InputLayer) -> tuple:
            try:
                # Keras 2.15.1 & below
                return layer.output_shape
            except AttributeError:
                return layer.output.shape  # pyright: ignore[reportAttributeAccessIssue]

        for i, layer in enumerate(model.layers):
            output_shape = _get_output_shape(layer)
            params[f"Layer-{i}"] = {
                "name": layer.name,
                "param": layer.count_params(),
                "output shape": output_shape,
            }
        params["Total # of weights"] = model.count_params()

        return params

    def _get_keras_info(self, model: KerasModel) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        try:
            return "keras", self._format_keras_params(model)
        except Exception:
            return None, None



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, List

from vectice.api.json.model_representation import ModelRepresentationOutput
from vectice.utils.common_utils import (
    convert_keys_to_snake_case,
    flatten_dict,
    process_versions_list_metrics_and_properties,
    remove_type_name,
    repr_class,
)
from vectice.utils.dataframe_utils import repr_list_as_pd_dataframe

if TYPE_CHECKING:
    from pandas import DataFrame

    from vectice.api.client import Client

_logger = logging.getLogger(__name__)


class ModelRepresentation:
    """Represents the metadata of a Vectice model.

    A Model Representation shows information about a specific model from the Vectice app. It makes it easier to get and read this information through the API.
    Those informations includes high level informations about versions inside the model.

    NOTE: **Hint**
        A model ID starts with 'MDL-XXX'. Retrieve the ID in the Vectice App, then use the ID with the following methods to get the model:
        ```connect.model('MDL-XXX')``` or ```connect.browse('MDL-XXX')```
        (see [Connection page](https://api-docs.vectice.com/reference/vectice/connection/#vectice.Connection.model)).

    Attributes:
        id (str): The unique identifier of the model.
        project_id (str): The identifier of the project to which the model belongs.
        name (str): The name of the model.
        description (str): The description of the model.
        type (str): The type of the model
        total_number_of_versions (int): The total number of versions belonging to this model.
        versions_status (Dict[str, Any]): Aggregated information about the statuses of model versions within the model (e.g., number of versions in production, staging...).
    """

    def __init__(
        self,
        output: ModelRepresentationOutput,
        client: "Client",
    ):
        self.id = output.id
        self.name = output.name
        self.type = output.type
        self.description = output.description
        self.version_status = output.version_stats
        self.total_number_of_versions = output.total_number_of_versions
        self.project_id = output.project_id

        self._last_version = output.version
        self._client = client

    def __repr__(self):
        return repr_class(self)

    def asdict(self) -> Dict[str, Any]:
        """Transform the ModelRepresentation into a organised dictionary.

        Returns:
            The object represented as a dictionary
        """
        return {
            "id": self.id,
            "name": self.name,
            "project_id": self.project_id,
            "description": self.description,
            "type": self.type,
            "version_status": self.version_status,
            "total_number_of_versions": self.total_number_of_versions,
        }

    def versions_list(
        self, number_of_versions: int = 30, as_dataframe: bool = True
    ) -> List[Dict[str, Any]] | DataFrame:
        """Retrieve the model versions list linked to the model. (Maximum of 100 versions).

        Parameters:
            number_of_versions (int): The number of versions to retrieve. Defaults to 30.
            as_dataframe (bool): If set to True, return type will be a pandas DataFrame for easier manipulation (requires pandas to be installed).
                                 If set to False, returns all model versions as a list of dictionaries.

        Returns:
            The model versions list as either a list of dictionaries or a pandas DataFrame.
        """
        if number_of_versions > 100:
            _logger.warning(
                "Only the first 100 versions will be retrieved. For additional versions, please contact your sales representative or email support@vectice.com"
            )
            number_of_versions = 100

        version_list = self._client.get_model_version_list(self.id, number_of_versions).list
        clean_version_list = remove_type_name(version_list)
        converted_version_list = [convert_keys_to_snake_case(versions) for versions in clean_version_list]
        processed_version_list = process_versions_list_metrics_and_properties(converted_version_list)

        if not as_dataframe:
            return processed_version_list

        for versions in processed_version_list:
            versions.update(flatten_dict(versions))
            versions["all_properties_dict"] = versions["properties"]
            versions["all_metrics_dict"] = versions["metrics"]
            del versions["properties"]
            del versions["metrics"]
        try:
            return repr_list_as_pd_dataframe(processed_version_list)
        except ModuleNotFoundError:
            _logger.info(
                "To display the list of versions as a DataFrame, pandas must be installed. Your list of versions will be in the format of a list of dictionaries."
            )
            return processed_version_list



from __future__ import annotations

import ast
import os
import re
from collections import OrderedDict
from typing import Any


class FilePathVisitor(ast.NodeVisitor):
    def __init__(self, is_graph_path: bool = False, is_dataset_path: bool = False):
        self.file_paths = set()
        self.dataset_file_paths = []
        self.is_graph_path = is_graph_path
        self.is_dataset_path = is_dataset_path

    def visit_Call(self, node: ast.Call) -> None:  # noqa: N802
        if isinstance(node.func, ast.Name) or isinstance(node.func, ast.Attribute):
            # Record function args, kwargs.
            path_var_and_path = {"variable": None, "path": None}
            supported_path = False
            try:
                func_name = node.func.attr  # pyright: ignore[reportAttributeAccessIssue]
                if self.is_graph_path:
                    supported_path = self._is_supported_graph_library(func_name)
                if self.is_dataset_path:
                    supported_path = self._is_supported_dataset_library(func_name)
            except AttributeError:
                pass

            args = node.args + node.keywords
            for arg in args:
                arg_value = self._get_arg_or_kwarg_val(arg)
                if arg_value and self._is_valid_file(arg_value) and supported_path:
                    if self.is_dataset_path:
                        path_var_and_path["path"] = arg_value
                        self.dataset_file_paths.append(path_var_and_path)
                    if self.is_graph_path:
                        self.file_paths.add(arg_value)

        self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign) -> None:
        path_var_and_path = {"variable": None, "path": None}

        try:
            if len(node.targets) == 1 and isinstance(
                node.targets[0], ast.Name
            ):  # pyright: ignore[reportAttributeAccessIssue]
                var_name = node.targets[0].id
                path_var_and_path["variable"] = var_name  # pyright: ignore[reportArgumentType]
                if isinstance(node.value, ast.Str):  # pyright: ignore[reportDeprecated]
                    path = node.value.s  # pyright: ignore[reportAttributeAccessIssue]
                    if self._is_valid_file(path):
                        if self.is_dataset_path:
                            path_var_and_path["path"] = path  # pyright: ignore[reportArgumentType]
                            self.dataset_file_paths.append(path_var_and_path)
                        if self.is_graph_path:
                            self.file_paths.add(path)
        except Exception:
            pass
        self.generic_visit(node)

    def _get_arg_or_kwarg_val(self, arg: Any) -> Any | None:
        try:
            # arg value
            if isinstance(arg.id, str):
                return arg.id  # pyright: ignore[reportAttributeAccessIssue]
        except AttributeError:
            pass
        try:
            # kwarg value if not a variable
            if isinstance(arg.value.value, str):
                return arg.value.value  # pyright: ignore[reportAttributeAccessIssue]
        except AttributeError:
            pass
        try:
            # kwarg value if it is a variable
            if isinstance(arg.value.id, str):
                return arg.value.id  # pyright: ignore[reportAttributeAccessIssue]
        except AttributeError:
            pass
        try:
            if isinstance(arg.value, str):
                return arg.value  # pyright: ignore[reportAttributeAccessIssue]
        except AttributeError:
            pass
        return None

    def _is_valid_file(self, path: str) -> bool:
        is_file = os.path.isfile(path)
        _, extension = os.path.splitext(path)
        if is_file or extension:
            return True
        return False

    def _is_supported_graph_library(self, func_name: str) -> bool:
        vectice_supported_graphs = [
            "savefig",
            "write_image",
        ]
        if func_name in vectice_supported_graphs:
            return True
        return False

    def _is_supported_dataset_library(self, func_name: str) -> bool:
        vectice_supported_datasets = [
            "read_csv",
        ]
        if func_name in vectice_supported_datasets:
            return True
        return False


class VariableVisitor(ast.NodeVisitor):
    def __init__(self, model_metrics: bool = False):
        self.variables: OrderedDict[str, None] = OrderedDict()
        self.processed_variables: set[str] = set()
        self.function_call_args: set[str] = set()
        self.function_call_kwargs: set[str] = set()
        self.variable_calls: set[str] = set()
        self.vectice_call_vars: set[str] = set()
        self.metric_variables: set[str] = set()
        self.model_metrics = model_metrics

    ##### Implement NodeVisitor methods

    def visit_Assign(self, node: ast.Assign) -> None:  # noqa: N802
        for target in node.targets:
            self._extract_variables_from_target(target)
        self.generic_visit(node)
        if self.model_metrics:
            self._extract_metric_vars(node)

    def visit_Name(self, node: ast.Name) -> None:  # noqa: N802
        var_name = node.id
        # get functions args and kwargs
        args_kwargs = self.function_call_args.union(self.function_call_kwargs)
        all_processed_vars = self.processed_variables.union(self.variable_calls)
        if var_name not in all_processed_vars and var_name not in args_kwargs:
            self.variables[var_name] = None
            self.processed_variables.add(var_name)
        self.generic_visit(node)

    def visit_Call(self, node: ast.Call) -> None:  # noqa: N802
        if isinstance(node.func, ast.Name) or isinstance(node.func, ast.Attribute):
            # Get variables that call functions
            if isinstance(node.func, ast.Attribute):
                try:
                    # Calls will fail
                    self.variable_calls.add(node.func.value.id)  # pyright: ignore[reportAttributeAccessIssue]
                except AttributeError:
                    pass
            # Record function args, kwargs.
            args = node.args + node.keywords
            for arg in args:
                arg_value = self._get_arg_or_kwarg_val(arg)
                if arg_value:
                    if hasattr(arg_value, "__iter__") and not isinstance(arg_value, str):
                        self.function_call_args = self.function_call_args.union(arg_value)
                    else:
                        self.function_call_args.add(arg_value)  # pyright: ignore[reportAttributeAccessIssue]

                func_name = self._get_function_name(node)
                if (
                    self._is_vectice_call_vars(func_name) or self._is_sklearn_search_call_vars(func_name)
                ) and arg_value:  # pyright: ignore[reportAttributeAccessIssue]
                    if hasattr(arg_value, "__iter__") and not isinstance(arg_value, str):
                        self.vectice_call_vars = self.vectice_call_vars.union(arg_value)
                    else:
                        self.vectice_call_vars.add(arg_value)
        self.generic_visit(node)

    def _get_function_name(self, node: ast.Call) -> str | None:
        try:
            if hasattr(node.func, "id"):
                return node.func.id  # pyright: ignore[reportAttributeAccessIssue]
            if hasattr(node.func, "value"):
                return node.func.value.id  # pyright: ignore[reportAttributeAccessIssue]
        except AttributeError:
            pass
        return None

    ##### VariableVisitor methods

    def _extract_variables_from_target(self, target: ast.expr) -> None:
        if isinstance(target, ast.Constant):
            return
        if isinstance(target, ast.Name):
            self.visit_Name(target)
        elif isinstance(target, ast.Tuple):
            for element in target.elts:
                if isinstance(element, ast.Name):
                    self.visit_Name(element)

    def _add_metric_variables(self, metric_variable: str, node: ast.Assign) -> None:
        try:
            function_call_name = node.value.func.id  # pyright: ignore[reportAttributeAccessIssue]
        except Exception:
            function_call_name = None

        is_metric_call = self._is_metric_call_vars(function_call_name) if function_call_name else False

        if is_metric_call:
            self.metric_variables.add(metric_variable)

    def _extract_metric_vars(self, node: ast.Assign) -> None:
        # Get the variable
        for target in node.targets:
            metric_variable = None
            if isinstance(target, ast.Name):
                metric_variable = target.id
            if metric_variable:
                self._add_metric_variables(metric_variable, node)

    def _get_arg_or_kwarg_val(self, arg: Any) -> Any | None:
        try:
            # arg value
            return arg.id
        except AttributeError:
            pass
        try:
            # kwarg value if not a variable
            return arg.value.value
        except AttributeError:
            pass
        try:
            # kwarg value if it is a variable
            return arg.value.id
        except AttributeError:
            pass
        try:
            # kwarg value if it is a list, list can contain multiple var types so we recursilvely get the arg/kwarg
            return {self._get_arg_or_kwarg_val(arg_value) for arg_value in arg.value.elts}
        except AttributeError:
            pass
        return None

    def _is_vectice_call_vars(self, func_name: str | None) -> bool:
        if not func_name:
            return False
        vectice_functions = [
            "NoResource",
            "Resource",
            "Table",
            "SnowflakeResource",
            "FileResource",
            "GCSResource",
            "S3Resource",
            "BigQueryResource",
            "DatabricksTableResource",
            "Dataset",
            "Model",
        ]
        if func_name in vectice_functions:
            return True
        return False

    def _is_sklearn_search_call_vars(self, func_name: str | None) -> bool:
        """Filter our sklearn search args & kwargs"""
        try:
            from sklearn.model_selection import _search  # pyright: ignore[reportPrivateUsage]
        except ImportError:
            return False

        if not func_name:
            return False
        sklearn_search_calls = dir(_search)
        return func_name in sklearn_search_calls

    def _is_metric_call_vars(self, func_name: str) -> bool:
        from sklearn.metrics import (
            _classification,  # pyright: ignore[reportPrivateUsage]
            _ranking,  # pyright: ignore[reportPrivateUsage]
            _regression,  # pyright: ignore[reportPrivateUsage]
            _scorer,  # pyright: ignore[reportPrivateUsage]
            cluster,  # pyright: ignore[reportPrivateUsage]
        )

        all_metrics = dir(_ranking) + dir(_scorer) + dir(cluster) + dir(_regression) + dir(_classification)

        if func_name in all_metrics:
            return True
        return False


def parse_comments(code: str) -> list[dict]:
    # Get all comments and variables
    comments_and_variables = r"##\s*(.*?)(?:$|\n)|(.+?)\s*=\s*.*?(?:$|\n)"
    all_comments_and_variables = []
    for idx, match in enumerate(re.findall(comments_and_variables, code)):
        comment, variable = match
        # Create placeholder for no variable found
        variable = variable if variable else f"variable_{idx}"
        # Remove empty string matches
        comment = comment if comment else None
        all_comments_and_variables.append({"variable": variable, "comment": comment})
    return all_comments_and_variables


def preprocess_code(code: str) -> str:
    # Remove or comment out lines with % or ! commands
    code = re.sub(r"^\s*%[^\n]*$", "#", code, flags=re.MULTILINE)
    code = re.sub(r"^\s*![^\n]*$", "#", code, flags=re.MULTILINE)
    # remove a notebooks helper function calls ?
    code = re.sub(r".*?\?$", "", code, flags=re.MULTILINE)

    return code



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from vectice.models.resource.metadata.column_metadata import Column, Size

T = TypeVar("T")


class DataFrameWrapper(ABC, Generic[T]):
    """Base class for dataframe wrappers.

    Use DataFrameWrapper subclasses to assign wrap DataFrames to generate statistics  The
    Vectice library supports a handful of common cases.  Additional
    cases are generally easy to supply by deriving from this base
    class.  In particular, subclasses must override this class'
    abstact methods (`capture_column_statistics()`, `capture_column_schema()`, `get_size()`).
    """

    def __init__(self, dataframe: T):
        self.dataframe = dataframe
        self.rows: int = 0

    @abstractmethod
    def get_size(self) -> Size:
        # self.rows to calculate
        pass

    @abstractmethod
    def capture_column_schema(self) -> list[Column]:
        pass

    @abstractmethod
    def capture_column_statistics(
        self,
        list_col_schema: list[Column],
        sample_rows_for_statistics: int,
        maximum_columns_for_statistics: int,
    ) -> list[Column]:
        pass

    def capture_columns(
        self,
        minimum_rows_for_statistics: int,
        sample_rows_for_statistics: int,
        maximum_columns_for_statistics: int,
        capture_schema_only: bool,
    ) -> list[Column]:
        list_schema = self.capture_column_schema()
        if self.rows >= minimum_rows_for_statistics and capture_schema_only is False:
            list_col_with_stats = self.capture_column_statistics(
                list_schema, sample_rows_for_statistics, maximum_columns_for_statistics
            )
            return list_col_with_stats

        return list_schema



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.json.dataset_representation import DatasetRepresentationOutput
from vectice.api.json.dataset_version_representation import DatasetVersionRepresentationOutput
from vectice.api.json.paged_response import PagedResponse
from vectice.utils.api_utils import PAGINATE_OUTPUT, get_page_input

if TYPE_CHECKING:
    from vectice.api.json.dataset_register import DatasetRegisterInput, DatasetRegisterOutput
    from vectice.api.json.iteration import IterationContextInput


_logger = logging.getLogger(__name__)

# TODO JobRun for lineages
_RETURNS = """
            datasetVersion {
                          vecticeId
                          name
                          origin {
                            id
                          }
                          dataSet {
                            name
                            projectId
                          }


            }
            useExistingDataset
            useExistingVersion
            __typename
            """

_RETURNS_DATASET = """
            vecticeId
            createdDate
            updatedDate
            name
            description
            type
            sourceOrigin
            versionCount
            lastVersion {
                vecticeId
                createdDate
                updatedDate
                name
                description
                __typename
            }
            project {
                vecticeId
            }
            __typename
"""

_BASE_DATASET_VERSION = """
            vecticeId
            createdDate
            updatedDate
            name
            description
            sourceOrigin
            properties {
                key
                value
            }
            __typename
"""

_LIST_VERSION = f"""
            {_BASE_DATASET_VERSION}
            datasetSources {{
                type
                usage
                size
                itemsCount
                columnsCount
                __typename
            }}
"""

_RETURNS_DATASET_VERSION = f"""
            {_BASE_DATASET_VERSION}
            dataSet {{
                vecticeId
                name
                project {{
                    vecticeId
                }}
                sourceOrigin
                type
                description
                versionCount
                lastVersion {{
                    vecticeId
                    __typename
                }}
            }}
            datasetSources {{
                usage
                itemsCount
                size
                columnsCount
            }}
"""


_RETURNS_EMPTY_DATASET_VERSION = """
            vecticeId
            __typename
"""

_RETURNS_PAGINATED_VERSION_LIST = PAGINATE_OUTPUT.format(_LIST_VERSION)


class GqlDatasetApi(GqlApi):
    def get_dataset(self, id: str) -> DatasetRepresentationOutput:
        variable_types = "$datasetId:VecticeId!"
        kw = "datasetId:$datasetId"
        variables = {"datasetId": id}
        query = GqlApi.build_query(
            gql_query="getDataset",
            variable_types=variable_types,
            returns=_RETURNS_DATASET,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: DatasetRepresentationOutput = Parser().parse_item(response["getDataset"])
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset", id)

    def get_dataset_version_list(self, id: str, size: int) -> PagedResponse[DatasetVersionRepresentationOutput]:
        gql_query = "getDatasetVersionsList"
        variable_types = "$datasetId:VecticeId!,$page:PageInput"
        kw = "datasetId:$datasetId,page:$page"
        variables = {
            "datasetId": id,
            "page": get_page_input(size=size),
        }
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_PAGINATED_VERSION_LIST,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: PagedResponse[DatasetVersionRepresentationOutput] = Parser().parse_paged_response(
                response[gql_query]
            )
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset_version", id)

    def get_dataset_version(self, id: str) -> DatasetVersionRepresentationOutput:
        variable_types = "$datasetVersionId:VecticeId!"
        kw = "datasetVersionId:$datasetVersionId"
        variables = {"datasetVersionId": id}
        query = GqlApi.build_query(
            gql_query="getDatasetVersion",
            variable_types=variable_types,
            returns=_RETURNS_DATASET_VERSION,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: DatasetVersionRepresentationOutput = Parser().parse_item(response["getDatasetVersion"])
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset_version", id)

    def get_dataset_version_has_columns_without_description(self, id: str) -> bool:
        variable_types = "$datasetVersionId:VecticeId!"
        kw = "datasetVersionId:$datasetVersionId"
        variables = {"datasetVersionId": id}
        query = GqlApi.build_query(
            gql_query="getDatasetVersion",
            variable_types=variable_types,
            returns="hasColumnWithoutDescription",
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            return self.execute(query_built, variables)["getDatasetVersion"]["hasColumnWithoutDescription"]
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset_version", id)

    def register_dataset(
        self,
        data: DatasetRegisterInput,
        iteration_context: IterationContextInput,
    ) -> DatasetRegisterOutput:
        variables: dict[str, Any] = {"iterationContext": iteration_context, "data": data}
        kw = "iterationContext:$iterationContext,data:$data"
        variable_types = "$iterationContext:IterationContextInput!,$data:DatasetRegisterInput!"

        query = GqlApi.build_query(
            gql_query="registerDataset",
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: DatasetRegisterOutput = Parser().parse_item(response["registerDataset"])
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset", "register dataset")

    def update_dataset_version(self, id: str, columns_description: list[dict[str, str]]):
        variable_types = "$datasetVersionId:VecticeId!,$data:DatasetVersionUpdateInput!"
        kw = "datasetVersionId:$datasetVersionId,data:$data"
        variables = {"datasetVersionId": id, "data": {"columnsDescription": columns_description}}
        query = GqlApi.build_query(
            gql_query="updateDatasetVersion",
            variable_types=variable_types,
            returns=_RETURNS_EMPTY_DATASET_VERSION,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            self.execute(query_built, variables)
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset_version", id)



from __future__ import annotations

import logging
import re
from textwrap import dedent
from typing import TYPE_CHECKING, ClassVar

from rich.table import Table

from vectice.api.http_error_handlers import (
    InvalidIdError,
    LastIterationNotWritableError,
    MultipleActiveIterationsError,
)
from vectice.api.json.iteration import IterationOutput, IterationStatus
from vectice.types.iteration import IterationInput
from vectice.utils.common_utils import temp_print
from vectice.utils.logging_utils import get_iteration_status
from vectice.utils.vectice_ids_regex import ITERATION_VID_REG

if TYPE_CHECKING:
    from vectice.api import Client
    from vectice.models import Phase
    from vectice.models.iteration import Iteration


_logger = logging.getLogger(__name__)


class PhaseService:
    __slots__: ClassVar[list[str]] = ["_client"]

    def __init__(self, client: Client):
        self._client: Client = client

    def list_iterations(
        self, phase: Phase, only_mine: bool = False, statuses: list[IterationStatus] | None = None
    ) -> None:
        """Prints a list of iterations belonging to the phase in a tabular format, limited to the last 10 items.

        Parameters:
            only_mine: Display only the iterations where the user is the owner.
            statuses: Filter iterations on specified statuses.

        Returns:
            None

        """
        iteration_outputs = self._client.list_iterations(phase.id, only_mine, statuses)
        rich_table = Table(expand=True, show_edge=False)

        rich_table.add_column("Id", justify="left", no_wrap=True, min_width=3, max_width=20)
        rich_table.add_column("Name", justify="left", no_wrap=True, min_width=5, max_width=20)
        rich_table.add_column("Status", justify="left", no_wrap=True, min_width=3, max_width=15)
        rich_table.add_column("Owner", justify="left", no_wrap=True, min_width=5, max_width=15)

        for iteration in iteration_outputs.list:
            rich_table.add_row(
                str(iteration.id),
                iteration.name,
                get_iteration_status(iteration.status, iteration.starred),
                iteration.ownername,
            )

        iteration_statuses_log = (
            " | ".join(list(map(lambda status: get_iteration_status(status), statuses)))
            if statuses is not None and len(statuses) > 0
            else None
        )
        status_log = f"with status [{iteration_statuses_log}] " if iteration_statuses_log is not None else ""
        only_mine_log = "You have" if only_mine is True else "There are"
        description = f"""{only_mine_log} {iteration_outputs.total} iterations {status_log}in the phase {phase.name!r} and a maximum of 10 iterations are displayed in the table below:"""
        link = dedent(
            f"""
        # For quick access to the list of iterations in the Vectice web app, visit:
        # {self._client.auth.api_base_url}/phase/{phase.id}/iterations"""
        ).lstrip()
        temp_print(description)
        temp_print(table=rich_table)
        temp_print(link)

    def create_or_get_current_iteration(
        self, phase: Phase, iteration: IterationInput | None = None
    ) -> Iteration | None:
        """Get or create an iteration.

        If your last updated iteration is writable (Not Started or In Progress), Vectice will return it.
        Otherwise, Vectice will create a new one and return it.
        If multiple writable iterations are found, Vectice will print a list of the iterations to complete or cancel.

        Returns:
            An iteration or None if Vectice could not determine which iteration to retrieve.

        Raises:
            VecticeException: When attempting to create an iteration but there isn't any step inside the phase.
        """
        try:
            retrieve_iteration_output = self._client.get_last_iteration(phase.id, iteration)
        except (LastIterationNotWritableError, MultipleActiveIterationsError) as e:
            _logger.warning(str(e.value))
            self.list_iterations(
                phase=phase, only_mine=True, statuses=[IterationStatus.NotStarted, IterationStatus.InProgress]
            )
            return None

        return self._build_iteration_from_output_and_log(
            phase=phase,
            iteration_output=retrieve_iteration_output.iteration,
            existing_iteration=retrieve_iteration_output.useExistingIteration,
        )

    def get_iteration_by_id_or_index(self, phase: Phase, index_or_id: int | str) -> Iteration:
        """Get an iteration.

        Fetch and return an iteration by index or id.

        Parameters:
            index: The index or id of an existing iteration.

        Returns:
            An iteration.

        Raises:
            InvalidIdError: When index is a string and not matching 'ITR-[int]'
            IterationIdError: Iteration with specified id does not exist.
            IterationIndexError: Iteration with specified index does not exist.
        """
        if isinstance(index_or_id, str):
            if re.search(ITERATION_VID_REG, index_or_id):
                iteration_output = self._client.get_iteration_by_id(index_or_id)
            else:
                raise InvalidIdError("iteration", index_or_id)
        else:
            iteration_output = self._client.get_iteration_by_index(phase_id=phase.id, index=index_or_id)

        return self._build_iteration_from_output_and_log(phase=phase, iteration_output=iteration_output)

    def create_iteration(self, phase: Phase, iteration: IterationInput | None = None) -> Iteration:
        """Create a new iteration.

        Create and return an iteration.

        Returns:
            An iteration.
        """
        from vectice.models.iteration import Iteration

        iteration_output = self._client.create_iteration(phase.id, iteration)
        logging_output = dedent(
            f"""
        New Iteration {iteration_output.name!r} created.

        For quick access to the Iteration in the Vectice web app, visit:
        {self._client.auth.api_base_url}/browse/iteration/{iteration_output.id}"""
        ).lstrip()
        _logger.info(logging_output)

        return Iteration(iteration_output, phase, self._client)

    def get_active_iteration_or_create(self, phase: Phase) -> Iteration:
        """Get latest active iteration or create one.

        Get or create and return an iteration.

        Parameters:
            phase: The iteration's phase.

        Returns:
            An iteration.
        """
        iteration_output = self._client.get_active_iteration_or_create(phase.id)
        return self._build_iteration_from_output_and_log(phase=phase, iteration_output=iteration_output)

    def _build_iteration_from_output_and_log(
        self, phase: Phase, iteration_output: IterationOutput, existing_iteration: bool = True
    ) -> Iteration:
        from vectice.models.iteration import Iteration

        iteration = Iteration(iteration_output, phase, self._client)
        base_log = dedent(
            f"Iteration {iteration.name!r} successfully {'retrieved' if existing_iteration else 'created'}."
        )
        if iteration_output.status in [IterationStatus.Completed, IterationStatus.Abandoned]:
            base_log += dedent(
                f"""
                WARN: Iteration is {iteration.status}."""
            )

        logging_output = dedent(
            f"""
                {base_log}

                For quick access to the Iteration in the Vectice web app, visit:
                {self._client.auth.api_base_url}/browse/iteration/{iteration.id}"""
        ).lstrip()
        _logger.info(logging_output)
        return iteration



from __future__ import annotations

from datetime import datetime


def read_nodejs_date(date_as_string: str | None) -> datetime | None:
    if date_as_string is None:
        return None
    return datetime.strptime(date_as_string, "%Y-%m-%dT%H:%M:%S.%f%z")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, List

from vectice.api.json.dataset_representation import DatasetRepresentationOutput
from vectice.utils.common_utils import (
    convert_keys_to_snake_case,
    flatten_dict,
    flatten_resources,
    process_versions_list_metrics_and_properties,
    remove_type_name,
    repr_class,
)
from vectice.utils.dataframe_utils import repr_list_as_pd_dataframe

if TYPE_CHECKING:
    from pandas import DataFrame

    from vectice.api.client import Client

_logger = logging.getLogger(__name__)


class DatasetRepresentation:
    """Represents the metadata of a Vectice dataset.

    A Dataset Representation shows information about a specific dataset from the Vectice app. It makes it easier to get and read this information through the API.

    NOTE: **Hint**
        A dataset ID starts with 'DTS-XXX'. Retrieve the ID in the Vectice App, then use the ID with the following methods to get the dataset:
        ```connect.dataset('DTS-XXX')``` or ```connect.browse('DTS-XXX')```
        (see [Connection page](https://api-docs.vectice.com/reference/vectice/connection/#vectice.Connection.dataset)).

    Attributes:
        id (str): The unique identifier of the dataset.
        project_id (str): The identifier of the project to which the dataset belongs.
        name (str): The name of the dataset.
        description (str): The description of the dataset.
        type (str): The type of the dataset.
        origin (str): The source origin of the dataset.
        total_number_of_versions (int): The total number of versions belonging to this dataset.
    """

    def __init__(
        self,
        output: DatasetRepresentationOutput,
        client: "Client",
    ):
        self.id = output.id
        self.project_id = output.project_id
        self.name = output.name
        self.type = output.type
        self.origin = output.origin
        self.description = output.description
        self.total_number_of_versions = output.total_number_of_versions
        self._last_version = output.version
        self._client = client

    def __repr__(self):
        return repr_class(self)

    def _asdict(self):
        return self.asdict()

    def asdict(self) -> Dict[str, Any]:
        """Transform the DatasetRepresentation into a organised dictionary.

        Returns:
            The object represented as a dictionary
        """
        return {
            "id": self.id,
            "name": self.name,
            "project_id": self.project_id,
            "description": self.description,
            "type": self.type,
            "origin": self.origin,
            "total_number_of_versions": self.total_number_of_versions,
        }

    def versions_list(
        self, number_of_versions: int = 30, as_dataframe: bool = True
    ) -> List[Dict[str, Any]] | DataFrame:
        """Retrieve the dataset versions list linked to the dataset. (Maximum of 100 versions).

        Parameters:
            number_of_versions (int): The number of versions to retrieve. Defaults to 30.
            as_dataframe (bool): If set to True, return type will be a pandas DataFrame for easier manipulation (requires pandas to be installed).
                                 If set to False, returns all dataset versions as a list of dictionaries.

        Returns:
            The dataset versions list as either a list of dictionaries or a pandas DataFrame.
        """
        if number_of_versions > 100:
            _logger.warning(
                "Only the first 100 versions will be retrieved. For additional versions, please contact your sales representative or email support@vectice.com"
            )
            number_of_versions = 100

        version_list = self._client.get_dataset_version_list(self.id, number_of_versions).list
        for version in version_list:
            version["datasetSources"] = flatten_resources(version.resources)
        clean_version_list = remove_type_name(version_list)
        converted_version_list = [convert_keys_to_snake_case(versions) for versions in clean_version_list]
        processed_version_list = process_versions_list_metrics_and_properties(converted_version_list)

        if not as_dataframe:
            return processed_version_list

        for versions in processed_version_list:
            versions.update(flatten_dict(versions))
            del versions["dataset_sources"]
            versions["all_properties_dict"] = versions["properties"]
            del versions["properties"]
        try:
            return repr_list_as_pd_dataframe(processed_version_list)
        except ModuleNotFoundError:
            _logger.info(
                "To display the list of versions as a DataFrame, pandas must be installed. Your list of versions will be in the format of a list of dictionaries."
            )
            return processed_version_list



from __future__ import annotations

import logging

from vectice.models.resource.base import Resource
from vectice.models.resource.metadata.base import DatasetSourceOrigin
from vectice.models.resource.metadata.dataframe_config import DataFrameType
from vectice.models.resource.metadata.db_metadata import DBMetadata, MetadataDB, TableType

_logger = logging.getLogger(__name__)

TDataFrameType = DataFrameType


class NoResource(Resource):
    """Wrap columnar data and its metadata in a local file.

    This resource wraps data that you have stored in a local file with optional metadata and versioning.
    You pass it as an argument of your Vectice Dataset wrapper before logging it to an iteration.

    ```python
    from vectice import FileResource

    resource = FileResource(paths="my/file/path")
    ```

    NOTE: **Vectice does not store your actual dataset. Instead, it stores your dataset's metadata, which is information about your dataset.
    These details are captured by resources tailored to the specific environment in use, such as: local (FileResource), Bigquery, S3, GCS...**
    """

    _origin = DatasetSourceOrigin.LOCAL.value

    def __init__(
        self,
        origin: str,
        dataframes: TDataFrameType | list[TDataFrameType] | None = None,
        capture_schema_only: bool = False,
        type: TableType = TableType.UNKNOWN,
        columns_description: dict[str, str] | None = None,
    ):
        """Initialize a file resource.

        Parameters:
            paths: The paths of the files to wrap.
            dataframes (Optional): The dataframes allowing vectice to optionally compute more metadata about this resource such as columns stats. (Support Pandas, Spark)
            capture_schema_only (Optional): A boolean parameter indicating whether to capture only the schema or both the schema and column statistics of the dataframes.
            type (Optional): The resource type.

        Examples:
            The following example shows how to wrap a CSV file
            called `iris.csv` in the current directory:

            ```python
            from vectice import FileResource
            iris_trainset = FileResource(paths="iris.csv")
            ```
        """
        super().__init__(
            paths="",
            dataframes=dataframes,
            capture_schema_only=capture_schema_only,
            columns_description=columns_description,
        )
        self._origin = origin
        self._type = type

    def _fetch_data(
        self,
    ) -> dict:
        return {}

    def _build_metadata(self) -> DBMetadata:
        dbs = []
        if self._dataframes:
            for dataframe in self._dataframes:
                dbs.append(
                    MetadataDB(
                        name="Dataframe",
                        uri="Dataframe",
                        dataframe=dataframe,
                        display_name="Dataframe",
                        capture_schema_only=self.capture_schema_only,
                        columns=[],
                        type=self._type,
                    )
                )

            metadata = DBMetadata(
                size=0,
                origin=self._origin,
                dbs=dbs,
            )
            return metadata
        raise ValueError("A dataframe is required.")



from __future__ import annotations

from vectice.models.resource.metadata.base import (
    DatasetSourceOrigin,
    DatasetSourceType,
    DatasetSourceUsage,
    DatasetType,
    Metadata,
)
from vectice.models.resource.metadata.column_metadata import Column, DBColumn
from vectice.models.resource.metadata.db_metadata import DBMetadata, MetadataDB
from vectice.models.resource.metadata.files_metadata import File, FilesMetadata

__all__ = [
    "DBMetadata",
    "Column",
    "DBColumn",
    "MetadataDB",
    "DatasetSourceOrigin",
    "DatasetSourceUsage",
    "DatasetSourceType",
    "DatasetType",
    "File",
    "FilesMetadata",
    "Metadata",
]



from typing import Any

from vectice.api.json.json_type import TJSON


class DatasetOutput(TJSON):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def project_id(self) -> int:
        return int(self["projectId"])


class DatasetVersionOutput(TJSON):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        if "dataSet" in self:
            self._dataset: DatasetOutput = DatasetOutput(**self["dataSet"])
        else:
            self._dataset = None  # type: ignore

    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def dataset(self) -> DatasetOutput:
        return self._dataset

    @property
    def lineage(self) -> dict:
        return self["origin"]

    @property
    def origin_id(self) -> int:
        return int(self["origin"]["id"])



from __future__ import annotations

from gql import gql

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.json.entity_file import EntityFileOutput

_RETURNS = """
            fileName
            __typename
            """


class GqlEntityFileApi(GqlApi):
    def get_entity_file_by_id(self, id: int) -> EntityFileOutput:
        gql_query = "getEntityFileById"
        variable_types = "$id:Float!"
        kw = "id:$id"
        variables = {"id": id}
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS, keyword_arguments=kw, query=True
        )
        query_built = gql(query)
        response = self.execute(query_built, variables)
        output: EntityFileOutput = Parser().parse_item(response[gql_query])
        return output



from __future__ import annotations


class VecticeError(Exception):
    def __init__(self, method_name: str, root_cause: Exception):
        self.method = method_name
        self.root_cause = root_cause

    def __str__(self):
        return "Error executing {method}: {cause}".format(method=self.method, cause=self.root_cause)



from __future__ import annotations

from enum import Enum


class DataDescription:
    """This class allows to describe a dataset and its columns."""

    def __init__(self, columns: list[DataDescriptionColumn], summary: str):
        """Initialize a data description.

        Parameters:
            columns: The described columns of the dataset.
            summary: A brief summary for the dataset.
        """
        self._columns = columns
        self._summary = summary

    @property
    def columns(self) -> list[DataDescriptionColumn]:
        """The columns description of the dataset.

        Returns:
            The columns description of the dataset.
        """
        return self._columns

    @columns.setter
    def columns(self, columns: list[DataDescriptionColumn]):
        """Set the columns description of the dataset.

        Parameters:
            columns: The columns description of the dataset.
        """
        self._columns = columns

    @property
    def summary(self) -> str:
        """Get the summary of the dataset.

        Returns:
            The summary.
        """
        return self._summary

    @summary.setter
    def summary(self, summary: str):
        """Set the summary of the dataset.

        Parameters:
            summary: The summary of the dataset.
        """
        self._summary = summary


class DataDescriptionColumn(dict):
    """Class used to describe dataset columns."""

    def __init__(self, column_name: str, column_description: str, column_data_type: ColumnDataType):
        """Initialize a data description column.

        Parameters:
            column_name: The name of the column.
            column_description: A brief description of the column.
            column_data_type: The type of data contained in the column.
        """
        super(DataDescriptionColumn, self).__init__()
        self.column_name = column_name
        self.column_description = column_description
        self.column_type = column_data_type


class ColumnDataType(Enum):
    # TODO: complete this list with other common data types
    OBJECT = "OBJECT"
    INT64 = "INT64"
    FLOAT64 = "FLOAT64"
    BOOL = "BOOL"
    DATETIME64 = "DATETIME64"
    TIMEDELTA = "TIMEDELTA"
    CATEGORY = "CATEGORY"
    OTHER = "OTHER"



from __future__ import annotations

from typing import TYPE_CHECKING

from vectice.autolog.asset_services.metric_service import MetricService
from vectice.autolog.asset_services.property_service import PropertyService
from vectice.autolog.asset_services.technique_service import TechniqueService
from vectice.autolog.model_library import ModelLibrary

if TYPE_CHECKING:
    from lightgbm.basic import Booster


class AutologLightgbmService(MetricService, PropertyService, TechniqueService):
    def __init__(self, key: str, asset: Booster, data: dict):
        self._asset = asset
        self._key = key

        super().__init__(cell_data=data)

    def get_asset(self):
        _, params = self._get_sklearn_or_xgboost_or_lgbm_info(self._asset)
        return {
            "variable": self._key,
            "model": self._asset,
            "library": ModelLibrary.LIGHTGBM,
            "technique": self._get_model_technique(self._asset, ModelLibrary.LIGHTGBM),
            "metrics": self._get_model_metrics(self._cell_data),
            "properties": params,
        }



from __future__ import annotations

from dataclasses import asdict, dataclass
from enum import Enum
from typing import Any, Dict

from vectice.utils.common_utils import convert_keys_to_camel_case


class ColumnCategoryType(Enum):
    """Enumeration of the column category types."""

    BOOLEAN = "BOOLEAN"
    NUMERICAL = "NUMERICAL"
    DATE = "DATE"
    TEXT = "TEXT"


@dataclass
class Quantiles:
    """Model a key-value pair of a quantiles of a numerical column.

    Parameters:
        q_min: min value
        q25: quantile 25%
        q50: median
        q75: quantile 75%
        q_max: max value
    """

    q_min: float
    q25: float
    q50: float
    q75: float
    q_max: float


@dataclass
class MostCommon:
    """Model a value-frequency pair of a most commun category in a column of dtypes str.

    Parameters:
        value: The key to identify the statistic.
        frequency: The value of the statistic.
    """

    value: str
    frequency: float


@dataclass
class NumericalStat:
    """Model a key-value pair of a column statistic for numerical.

    Parameters:
        missing: The percentage of missing value between O and 1
        mean: mean of the column
        std_deviation: standard deviation of the column
        quantiles: modelise a Quantiles object
    """

    missing: float
    mean: float
    std_deviation: float
    quantiles: Quantiles


@dataclass
class DateStat:
    """Model a key-value pair of date column statistic.

    Parameters:
        missing: The percentage of missing value between O and 1
        minimum: the first date
        mean: the average date
        median: the median date
        maximum: the last date
    """

    missing: float
    minimum: str
    mean: str
    median: str
    maximum: str


@dataclass
class TextStat:
    """Model a key-value pair of a column statistic.

    Parameters:
        missing: The percentage of missing value between O and 1
        unique: The value of the statistic.
        most_commons: a list of MostCommon objects with value-frequency
    """

    missing: float
    unique: float
    most_commons: list[MostCommon]


@dataclass
class BooleanStat:
    """Model a key-value pair of a column statistic.

    Parameters:
        missing: The percentage of missing value between O and 1
        true: The percentage of True value between O and 1
        false: The percentage of False value between O and 1
    """

    missing: float
    true: float
    false: float


class Size:
    """Model size info of a dataframe."""

    def __init__(self, rows: int, columns: int):
        """Initialize the metadata of the DataFrame.

        Parameters:
            rows  The number of rows.
            columns : The number of columns.
        """
        self.rows = rows
        self.columns = columns

    def asdict(self) -> Dict[str, int | float | None]:
        return {"rowsNumber": self.rows, "columnsNumber": self.columns}


class Column:
    """Model a column of a dataset."""

    def __init__(
        self,
        name: str,
        data_type: str,
        stats: BooleanStat | TextStat | NumericalStat | DateStat | None = None,
        category_type: ColumnCategoryType | None = None,
    ):
        """Initialize a column.

        Parameters:
            name: The name of the column.
            data_type: The type of the data contained in the column.
            stats: Additional statistics about the column.
            category_type: Column category type.
        """
        self.name = name
        self.data_type = data_type
        self.stats = stats
        self.category_type = category_type

    def asdict(self) -> Dict[str, Any]:
        obj = {"name": self.name, "dataType": self.data_type, "statistics": {}}
        if self.category_type:
            obj["categoryType"] = self.category_type.value
            category_value = self.category_type.value.lower()
            if self.stats is not None:
                obj["statistics"] = {category_value: convert_keys_to_camel_case(asdict(self.stats))}

        return obj


class DBColumn(Column):
    """Model a column of a dataset, like a database column."""

    def __init__(
        self,
        name: str,
        data_type: str,
        is_unique: bool | None = None,
        nullable: bool | None = None,
        is_private_key: bool | None = None,
        is_foreign_key: bool | None = None,
        stats: BooleanStat | TextStat | NumericalStat | DateStat | None = None,
    ):
        """Initialize a column.

        Parameters:
            name: The name of the column.
            data_type: The type of the data contained in the column.
            is_unique: If the column uniquely defines a record.
            nullable: If the column can contain null value.
            is_private_key: If the column uniquely defines a record,
                individually or with other columns (can be null).
            is_foreign_key: If the column refers to another one,
                individually or with other columns (cannot be null).
            stats: Additional statistics about the column.
        """
        super().__init__(name, data_type, stats)
        self.is_unique = is_unique
        self.nullable = nullable
        self.is_private_key = is_private_key
        self.is_foreign_key = is_foreign_key

    def asdict(self) -> dict:
        return {
            **super().asdict(),
            "isUnique": self.is_unique,
            "nullable": self.nullable,
            "isPK": self.is_private_key,
            "isFK": self.is_foreign_key,
        }



from __future__ import annotations

from dataclasses import dataclass, field
from enum import EnumMeta
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from datetime import datetime


class FileMetadataType(EnumMeta):
    """Enumeration of the supported types of resources."""

    Folder = "Folder"
    CsvFile = "CsvFile"
    ImageFile = "ImageFile"
    ExcelFile = "ExcelFile"
    TextFile = "TextFile"
    MdFile = "MdFile"
    DataSet = "DataSet"
    DataTable = "DataTable"
    File = "File"
    Notebook = "Notebook"


@dataclass
class FileMetadata:
    name: str | None = None
    id: str | None = None
    parentId: str | None = None
    path: str | None = None
    type: FileMetadataType | str | None = None
    isFolder: bool | None = False
    children: list[FileMetadata] = field(default_factory=list)
    size: int | None = 0
    uri: str | None = None
    generation: str | None = None
    digest: str | None = None
    itemCreatedDate: datetime | None = None
    itemUpdatedDate: datetime | None = None

    def append(self, child: FileMetadata):
        return self.children.append(child)



from __future__ import annotations

from abc import abstractmethod
from importlib.util import find_spec
from typing import Any, Protocol

from vectice.api.http_error_handlers import VecticeException
from vectice.autolog.asset_services import (
    AutologCatboostService,
    AutologKerasService,
    AutologLightgbmService,
    AutologPandasService,
    AutologPysparkService,
    AutologPytorchService,
    AutologSklearnService,
    AutologStatsModelWrapperService,
    AutologVecticeAssetService,
    VecticeObjectClasses,
)


class IAutologService(Protocol):
    @abstractmethod
    def get_asset(self) -> dict[str, Any] | None: ...


class AssetFactory:
    @staticmethod
    def get_asset_service(key: str, asset: Any, data: dict) -> IAutologService:
        is_pandas = find_spec("pandas") is not None
        is_pyspark = find_spec("pyspark") is not None
        is_lgbm = find_spec("lightgbm") is not None
        is_sklearn = find_spec("sklearn") is not None
        is_catboost = find_spec("catboost") is not None
        is_keras = find_spec("keras") is not None
        is_statsmodels = find_spec("statsmodels") is not None
        is_pytorch = find_spec("torch") is not None

        if is_pandas:
            from pandas import DataFrame

            if isinstance(asset, DataFrame):
                return AutologPandasService(key, asset)

        if is_pyspark:
            from pyspark.sql import DataFrame as SparkDF

            if isinstance(asset, SparkDF):
                return AutologPysparkService(key, asset)

        if is_lgbm:
            from lightgbm.basic import Booster

            if isinstance(asset, Booster):
                return AutologLightgbmService(key, asset, data)

        if is_catboost:
            from catboost.core import CatBoost

            if isinstance(asset, CatBoost):
                return AutologCatboostService(key, asset, data)

        if is_keras:
            from keras.models import Model as KerasModel  # type: ignore[reportMissingImports]

            if isinstance(asset, KerasModel):
                return AutologKerasService(key, asset, data)

        if is_pytorch:
            from torch.nn import Module

            if isinstance(asset, Module):
                return AutologPytorchService(key, asset, data)

        if isinstance(asset, VecticeObjectClasses):
            return AutologVecticeAssetService(key, asset)  # type: ignore[reportArgumentType]

        if is_statsmodels:
            from statsmodels.base.wrapper import ResultsWrapper

            if isinstance(asset, ResultsWrapper):
                return AutologStatsModelWrapperService(key, asset, data)

        if is_sklearn:
            return AutologSklearnService(key, asset, data)

        raise VecticeException(f"Asset {asset} of type ({type(asset)!r}) not handled")



from __future__ import annotations

import json
from typing import TYPE_CHECKING

from sklearn.base import BaseEstimator
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline

if TYPE_CHECKING:
    from vectice.autolog.autolog_class import ModelTypes


def _get_non_default_params(step: BaseEstimator):
    """Return parameters that differ from the default values."""
    default_params = type(step)().get_params(deep=False)
    current_params = step.get_params(deep=False)
    non_default_params = {}
    for key, value in current_params.items():
        if key not in default_params or default_params[key] != value:
            non_default_params[key] = value
    return non_default_params


def get_params_recursive(step: BaseEstimator):
    """Return parameters in a serializable format."""
    if hasattr(step, "get_params"):
        params = _get_non_default_params(step)
        serializable_params = {}
        for key, value in params.items():
            if hasattr(value, "get_params"):
                serializable_params[key] = get_params_recursive(value)
            else:
                try:
                    value = json.dumps(value)  # Test if the value is serializable
                    serializable_params[key] = value
                except TypeError:
                    serializable_params[key] = str(value)
        return serializable_params
    else:
        return str(step)


def process_step(step: BaseEstimator):
    """Process each step in the pipeline, handling nested pipelines and column transformers."""
    if isinstance(step, Pipeline):
        steps = []
        for name, sub_step in step.steps:
            steps.append({"name": name, "type": type(sub_step).__name__, "parameters": process_step(sub_step)})
        return {"type": "Pipeline", "steps": steps}
    elif isinstance(step, ColumnTransformer):
        transformers = []
        for name, transformer, columns in step.transformers:
            transformers.append(
                {
                    "name": name,
                    "type": type(transformer).__name__,
                    "columns": columns,
                    "parameters": process_step(transformer) if hasattr(transformer, "get_params") else str(transformer),
                }
            )
        return {"type": "ColumnTransformer", "transformers": transformers}
    else:
        return {"type": type(step).__name__, "parameters": get_params_recursive(step)}


def pipeline_to_json(pipeline: ModelTypes) -> str | None:
    """Convert the entire pipeline to a JSON-compatible dictionary."""
    try:
        if isinstance(pipeline, Pipeline):
            pipeline_dict = process_step(pipeline)
            return json.dumps(pipeline_dict, indent=4)
    except Exception:
        pass
    return None



from __future__ import annotations

import io
import logging
import os
from functools import reduce
from typing import TYPE_CHECKING, BinaryIO, List

from vectice.api.json.model_version import ModelVersionOutput
from vectice.models.attachment import TFormattedAttachment
from vectice.models.table import Table
from vectice.types.version import TVersion
from vectice.utils.common_utils import capture_unique_attachments

if TYPE_CHECKING:
    from vectice.api import Client

_logger = logging.getLogger(__name__)

FILE_PATH_DOES_NOT_EXIST_ERROR_MESSAGE = "The file path '%s' is not valid. The file does not exist."


class AttachmentContainer:
    def __init__(self, version: TVersion, client: Client):
        self._client = client
        self._version = version
        self._id = version.id
        self._name = version.name

    @property
    def version(self):
        return self._version

    @property
    def name(self):
        return self._name

    @property
    def id(self):
        return self._id

    def upsert_attachments(self, attachments: list[TFormattedAttachment]):
        """Add an attachment or set of attachments to the entity.

        Parameters:
            attachments: The paths of the attachment(s).

        Returns:
            A list of dictionaries describing the attachments.
        """

        def _upsert_files(files: list[str]) -> list[str]:
            if len(files) > 0:
                files_attachments = self._add_files_to_attachments(files)
                try:
                    return list(
                        map(
                            lambda curr: curr.fileName,
                            self._client.upsert_version_attachments(files_attachments, self._version),
                        )
                    )
                finally:
                    self._close_attached_files(files_attachments)
            return []

        def _upsert_tables(tables: list[Table]) -> list[str]:
            return (
                list(map(lambda curr: curr["name"], self._client.upsert_version_tables(tables, self._version)))
                if len(tables) > 0
                else []
            )

        files, tables = reduce(capture_unique_attachments, attachments, ([], []))
        return [*_upsert_files(files), *_upsert_tables(tables)]

    def _add_files_to_attachments(self, file_paths: list[str]) -> list[tuple[str, tuple[str, BinaryIO]]]:
        attachments: List[tuple[str, tuple[str, BinaryIO]]] = []
        for file_path in file_paths:
            if not os.path.exists(file_path):
                raise ValueError(FILE_PATH_DOES_NOT_EXIST_ERROR_MESSAGE % file_path)
            curr_file = ("file", (file_path, open(file_path, "rb")))
            attachments.append(curr_file)
        return attachments

    def _close_attached_files(self, attachments: list[tuple[str, tuple[str, BinaryIO]]]) -> None:
        for attachment in attachments:
            attachment[1][1].close()

    def add_serialized_model(self, model_type: str, model_content: bytes):
        if isinstance(self._version, ModelVersionOutput):
            self._client.create_model_predictor(model_type, io.BytesIO(model_content), self._version)



from __future__ import annotations

from typing import Literal

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi
from vectice.api.json.iteration import IterationStepArtifactEntityMetadataInput
from vectice.models.table import Table
from vectice.utils.dataframe_utils import transform_table_to_metadata_dict

_RETURNS = """
            name
            __typename
            """


class GqlAttachmentApi(GqlApi):
    def upsert(
        self, entity_id: str, entity_type: Literal["DATASET_VERSION"] | Literal["MODEL_VERSION"], tables: list[Table]
    ):
        def serialize_table(asset: Table):
            data = transform_table_to_metadata_dict(asset)
            return IterationStepArtifactEntityMetadataInput(name=asset.name, content={"data": data})

        gql_query = "upsertEntityMetadata"
        variable_types = "$entityId:VecticeId!,$entityType:EntityMetadataTypeEnum!,$data:[EntityMetadataCreateInput!]!"
        kw = "entityId:$entityId,entityType:$entityType,data:$data"
        variables = {"entityId": entity_id, "entityType": entity_type, "data": list(map(serialize_table, tables))}
        query = GqlApi.build_query(
            gql_query="upsertEntityMetadata",
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            return self.execute(query_built, variables)[gql_query]
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "attachment", "upsert")

    def attach_file_to_lineage(self, lineage_id: int, entity_file_id: int):
        gql_query = "attachFileToLineage"
        variable_types = "$lineageId:Float!,$entityFileId:Float!"
        kw = "lineageId:$lineageId,entityFileId:$entityFileId"
        variables = {"lineageId": lineage_id, "entityFileId": entity_file_id}
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=None,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            return self.execute(query_built, variables)[gql_query]
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "attachment", "attach")



from __future__ import annotations

DEFAULT_REG = r"\d+$"
WORKSPACE_VID_REG = f"WSP-{DEFAULT_REG}"
PROJECT_VID_REG = f"PRJ-{DEFAULT_REG}"
PHASE_VID_REG = f"PHA-{DEFAULT_REG}"
ITERATION_VID_REG = f"ITR-{DEFAULT_REG}"
DATASET_VID_REG = f"DTS-{DEFAULT_REG}"
DATASET_VERSION_VID_REG = f"DTV-{DEFAULT_REG}"
MODEL_VID_REG = f"MDL-{DEFAULT_REG}"
MODEL_VERSION_VID_REG = f"MDV-{DEFAULT_REG}"



from io import IOBase
from typing import Any, Union

from PIL.Image import Image

from vectice.utils.common_utils import check_file_path, check_image_path


def is_image_or_file(value: Any) -> Union[Image, str, IOBase, None]:
    return is_pil_image(value) or is_binary(value) or is_existing_image_path(value) or is_existing_file_path(value)


def is_pil_image(value: Any) -> Union[Image, None]:
    return value if isinstance(value, Image) else None


def is_binary(value: Any) -> Union[IOBase, None]:
    return value if isinstance(value, IOBase) else None


def is_existing_image_path(value: Any):
    return value if isinstance(value, str) and check_image_path(value) else None


def is_existing_file_path(value: Any):
    return value if isinstance(value, str) and check_file_path(value) else None



from __future__ import annotations

from typing import Any

from vectice.api.json.json_type import TJSON
from vectice.api.json.workspace import WorkspaceOutput
from vectice.types.project import ProjectInput


class ProjectCreateInput(TJSON):
    def __init__(self, project: ProjectInput):
        self["name"] = project["name"]
        if "template" in project:
            self["templateName"] = project["template"]
        if "copy_project_id" in project:
            self["duplicateProjectId"] = project["copy_project_id"]
        if "description" in project:
            self["description"] = project["description"]


class ProjectOutput(TJSON):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        if "workspace" in self:
            self._workspace: WorkspaceOutput = WorkspaceOutput(**self["workspace"])

    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str | None:
        if "description" in self and self["description"] is not None:
            return str(self["description"])
        else:
            return None

    @property
    def workspace(self) -> WorkspaceOutput:
        return self._workspace



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from gql import gql

from vectice.api.json.paged_response import PagedResponse
from vectice.utils.api_utils import PAGINATE_OUTPUT, get_page_input
from vectice.utils.vectice_ids_regex import WORKSPACE_VID_REG

if TYPE_CHECKING:
    from vectice.api.json import WorkspaceOutput

from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser

_RETURNS = """
            name
            description
            vecticeId
            __typename
"""

_RETURNS_PAGE = PAGINATE_OUTPUT.format(_RETURNS)


class WorkspaceApi(GqlApi):
    def get_workspace(self, workspace: str) -> WorkspaceOutput:
        search_vectice_id = re.search(WORKSPACE_VID_REG, workspace)
        if search_vectice_id:
            gql_query = "getWorkspaceById"
            variable_types = "$workspaceId:VecticeId!"
            variables = {"workspaceId": workspace}
            kw = "workspaceId:$workspaceId"
        else:
            gql_query = "getWorkspaceByName"
            variable_types = "$name:String!"
            variables = {"name": workspace}
            kw = "name:$name"
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS, keyword_arguments=kw, query=True
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            workspace_output: WorkspaceOutput = Parser().parse_item(response[gql_query])
            return workspace_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "workspace", workspace)

    def list_workspaces(self) -> PagedResponse[WorkspaceOutput]:
        gql_query = "getUserWorkspaceList"
        variable_types = "$page: PageInput!"
        variables = {"page": get_page_input()}
        kw = "page:$page"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_PAGE,
            keyword_arguments=kw,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            workspace_output: PagedResponse[WorkspaceOutput] = Parser().parse_paged_response(response[gql_query])
            return workspace_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "workspaces", "list")



from __future__ import annotations

import logging

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.json.public_config import PublicConfigOutput

_logger = logging.getLogger(__name__)

_RETURNS = """
    versions {
        version
        artifactName
    }
    __typename
"""


class VersionApi(GqlApi):
    def get_public_config(self) -> PublicConfigOutput:
        gql_query = "getPublicConfig"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=None,
            returns=_RETURNS,
            keyword_arguments=None,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, None)
            return Parser().parse_item(response[gql_query])
        except TransportQueryError as error:
            raise TransportQueryError("Impossible to get the public configuration.") from error



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyspark.pandas.frame import DataFrame as PysparkPandasDF

from datetime import datetime

from vectice.models.resource.metadata.column_metadata import (
    DateStat,
)
from vectice.models.resource.metadata.df_wrapper_pandas_default import DataFrameDefaultPandasWrapper
from vectice.models.resource.metadata.pyspark_pandas_dataframe_typing import Series

_logger = logging.getLogger(__name__)


try:
    from pyspark.sql import SparkSession
    from pyspark.sql.functions import avg, expr

    spark_imported = True
except ImportError:
    spark_imported = False

    from vectice.models.resource.metadata.pyspark_typing import avg, expr

    pass


class PysparkPandasDFWrapper(DataFrameDefaultPandasWrapper):
    def __init__(self, dataframe: PysparkPandasDF):
        if spark_imported is False:
            raise ImportError("Pyspark is not installed.")
        super().__init__(dataframe)
        self.spark = SparkSession.builder.getOrCreate()  # type: ignore @see: https://github.com/microsoft/pylance-release/issues/4577

    def __compute_date_column_statistics__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, series: Series
    ) -> DateStat:
        """Parse a dataframe series and return statistics about it.

        The computed statistics are:
        - the first date
        - the mean
        - the median
        - the last date
        - the count missing value in %
        Parameters:
            series: The pandas series to get information from.

        Returns:
            A DateStat object containing the above statistics.
        """
        # Convert to datetime since mean is not supported for non datetime pandas object such as dbdates
        series_time = series.astype("datetime64[ns]")
        min_date = series_time.min().isoformat()
        max_date = series_time.max().isoformat()
        missing = series.isnull().sum() / len(series_time)

        _temp = series_time.to_frame().to_spark()
        col_name = _temp.columns[0]

        df = _temp.withColumn(col_name, _temp[col_name].cast("timestamp"))
        mean = datetime.isoformat(df.agg(avg(col_name).cast("timestamp").alias("avg_created_datetime")).collect()[0][0])

        df = _temp.withColumn(col_name, expr(f"UNIX_TIMESTAMP(`{col_name}`, 'yyyy-MM-dd')"))
        median = df.approxQuantile(col_name, [0.5], 0.0)[0]
        median_iso = datetime.strptime(
            self.spark.sql(f"SELECT from_unixtime({median!s}, 'yyyy-MM-dd')").collect()[0][0], "%Y-%m-%d"
        ).isoformat()

        return DateStat(
            missing=float(missing), minimum=str(min_date), mean=str(mean), median=str(median_iso), maximum=str(max_date)
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from vectice.models.resource.metadata.db_metadata import TableType

if TYPE_CHECKING:
    from pandas import DataFrame


class AutologPandasService:
    def __init__(self, key: str, asset: DataFrame):
        self._asset = asset
        self._key = key

    def get_asset(self):
        return {
            "variable": self._key,
            "dataframe": self._asset,
            "type": TableType.PANDAS,
        }



from __future__ import annotations

import ast
import math
from functools import reduce
from typing import TYPE_CHECKING, Any

from vectice.utils.code_parser import VariableVisitor, preprocess_code

if TYPE_CHECKING:
    from keras.models import Model as KerasModel  # type: ignore[reportMissingImports]
    from numpy import float_
    from statsmodels.base.wrapper import ResultsWrapper


# basic metrics
_autolog_metric_allowlist = [
    "aic",
    "bic",
    "centered_tss",
    "condition_number",
    "df_model",
    "df_resid",
    "ess",
    "f_pvalue",
    "fvalue",
    "llf",
    "mse_model",
    "mse_resid",
    "mse_total",
    "rsquared",
    "rsquared_adj",
    "scale",
    "ssr",
    "uncentered_tss",
]


class MetricService:
    def __init__(self, cell_data: dict):
        self._cell_data = cell_data
        self._model_cell = None

    def _get_model_metrics(self, data: dict) -> dict[str, Any]:
        # TODO mix of models ?
        cell_content = data["cell"]
        variables = data["variables"]

        if not cell_content:
            return {}
        # Get model cell content for metrics
        self._model_cell = preprocess_code(cell_content)

        tree = ast.parse(self._model_cell)
        visitor = VariableVisitor(True)
        visitor.visit(tree)

        metrics = list(visitor.metric_variables)

        return reduce(
            lambda identified_metrics, key: (
                {**identified_metrics, key: variables[key]}
                if key in metrics and isinstance(variables[key], (int, float))
                else identified_metrics
            ),
            variables.keys(),
            {},
        )

    def _get_keras_training_metrics(self, model: KerasModel) -> dict[str, float]:
        try:
            return {
                str(key)
                + "_train": float(model.get_metrics_result()[key].numpy())  # pyright: ignore[reportGeneralTypeIssues]
                for key in model.get_metrics_result().keys()  # pyright: ignore[reportGeneralTypeIssues]
            }
        except Exception:
            return {}

    def _get_statsmodels_metrics(self, model: ResultsWrapper):
        try:
            # statsmodels can function without numpy
            import numpy

            has_numpy = True
        except ImportError:
            has_numpy = False

        def _convert_metric(metric_value: float | float_) -> float | None:
            if metric_value is numpy.float_:  # pyright: ignore[reportPossiblyUnboundVariable]
                metric_value = metric_value.item()

            if math.isnan(float(metric_value)):
                return None
            return round(float(metric_value), 4)

        result_metrics = {}
        for metric in _autolog_metric_allowlist:
            try:
                if hasattr(model, metric):
                    metric_value = getattr(model, metric)
                    if has_numpy:
                        # handle numpy floats
                        converted_metric = _convert_metric(metric_value)
                    else:
                        # simple conversion
                        converted_metric = None if math.isnan(float(metric_value)) else round(float(metric_value), 4)

                    if converted_metric:
                        result_metrics[metric] = _convert_metric(metric_value)
            except Exception:
                pass
        return result_metrics



from __future__ import annotations

from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
from vectice.models.representation.model_representation import ModelRepresentation
from vectice.models.representation.model_version_representation import ModelVersionRepresentation

__all__ = ["ModelVersionRepresentation", "DatasetVersionRepresentation", "ModelRepresentation"]



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from vectice.api.json.phase import PhaseOutput, PhaseStatus
from vectice.services import PhaseService

if TYPE_CHECKING:
    from vectice import Connection
    from vectice.api import Client
    from vectice.api.json.iteration import IterationStatus
    from vectice.models import Project, Workspace
    from vectice.models.iteration import Iteration


class Phase:
    """Represent a Vectice phase.

    Phases reflect the real-life phases of the project lifecycle
    (i.e., Business Understanding, Data Preparation, Modeling,
    Deployment, etc.).

    Phases let you document the goals, assets, and outcomes along with
    the status to organize the project, enforce best practices, allow
    consistency, and capture knowledge.

    To get a project's phase:

    ```python
    phase = project.phase("Business Understanding")
    ```

    Iterations can then be created for this phase.

    ```python
    my_origin_dataset = ...
    iteration = phase.create_iteration()
    iteration.log(my_origin_dataset)
    ```

    NOTE: **Phases are created in the Vectice App,
    iterations are created from the Vectice python API.**

    See the documentation of [Iterations][vectice.models.Iteration]
    for more information about iterations.
    """

    __slots__: ClassVar[list[str]] = [
        "_id",
        "_project",
        "_name",
        "_index",
        "_status",
        "_current_iteration",
        "_pointers",
        "_client",
        "_phase_service",
    ]

    def __init__(
        self,
        output: PhaseOutput,
        project: Project,
        client: Client,
    ):
        self._id = output.id
        self._project = project
        self._name = output.name
        self._index = output.index
        self._status = output.status
        self._client = client
        # TODO it seems useless, remove it
        self._current_iteration: Iteration | None = None
        # TODO this must be injected at phase creation so we have only one service instance, keeping like this for a 1st implementation
        self._phase_service = PhaseService(self._client)

    def __repr__(self):
        return f"Phase (name='{self.name}', id={self.id}, status='{self.status.name}')"

    def __eq__(self, other: object):
        if not isinstance(other, Phase):
            return NotImplemented
        return self.id == other.id

    @property
    def id(self) -> str:
        """The phase's id.

        Returns:
            The phase's id.
        """
        return self._id

    @id.setter
    def id(self, phase_id: str):
        """Set the phase's id.

        Parameters:
            phase_id: The phase id to set.
        """
        self._id = phase_id

    @property
    def name(self) -> str:
        """The phase's name.

        Returns:
            The phase's name.
        """
        return self._name

    @property
    def index(self) -> int:
        """The phase's index.

        Returns:
            The phase's index.
        """
        return self._index

    @property
    def status(self) -> PhaseStatus:
        """The phase's status.

        Returns:
            The phase's status.
        """
        return self._status

    @property
    def properties(self) -> dict[str, str | int]:
        """The phase's name, id, and index.

        Returns:
            A dictionary containing the `name`, `id`, and `index` items.
        """
        return {"name": self.name, "id": self.id, "index": self.index}

    @property
    def connection(self) -> Connection:
        """The connection to which this phase belongs.

        Returns:
            The connection to which this phase belongs.
        """
        return self._project.connection

    @property
    def workspace(self) -> Workspace:
        """The workspace to which this phase belongs.

        Returns:
            The workspace to which this phase belongs.
        """
        return self._project.workspace

    @property
    def project(self) -> Project:
        """The project to which this phase belongs.

        Returns:
            The project to which this phase belongs.
        """
        return self._project

    def list_iterations(self, only_mine: bool = False, statuses: list[IterationStatus] | None = None) -> None:
        """Prints a list of iterations belonging to the phase in a tabular format, limited to the last 10 items.

        Parameters:
            only_mine: Display only the iterations where the user is the owner.
            statuses: Filter iterations on specified statuses.

        Returns:
            None

        """
        return self._phase_service.list_iterations(phase=self, only_mine=only_mine, statuses=statuses)

    def create_or_get_current_iteration(self) -> Iteration | None:
        """Get or create an iteration.

        If your last updated iteration is writable (Not Started or In Progress), Vectice will return it.
        Otherwise, Vectice will create a new one and return it.
        If multiple writable iterations are found, Vectice will print a list of the iterations to complete or cancel.

        Returns:
            An iteration or None if Vectice could not determine which iteration to retrieve.
        """
        iter = self._phase_service.create_or_get_current_iteration(self)
        self._current_iteration = iter
        return self._current_iteration

    def iteration(self, index: int | str) -> Iteration:
        """Get an iteration.

        Fetch and return an iteration by index or id.

        Parameters:
            index: The index or id of an existing iteration.

        Returns:
            An iteration.

        Raises:
            InvalidIdError: When index is a string and not matching 'ITR-[int]'
            IterationIdError: Iteration with specified id does not exist.
            IterationIndexError: Iteration with specified index does not exist.
        """
        iteration = self._phase_service.get_iteration_by_id_or_index(phase=self, index_or_id=index)
        self._current_iteration = iteration
        return self._current_iteration

    def create_iteration(self, name: str | None = None, description: str | None = None) -> Iteration:
        """Create a new iteration.

        Create and return an iteration.

        Parameters:
            name: The iteration's name, default being its index.
            description: The iteration's description.

        Returns:
            An iteration.
        """
        if name is not None:
            self._client.assert_feature_flag_or_raise("rename-iteration-from-the-api")
        if description is not None:
            self._client.assert_feature_flag_or_raise("iteration-description")

        iter = self._phase_service.create_iteration(self, {"name": name, "description": description})
        self._current_iteration = iter
        return self._current_iteration
